//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// デバッガ (内部用)
//

#pragma once

#include "branchhistory.h"
#include "bus.h"
#include "monitor.h"
#include "textscreen.h"
#include "thread.h"
#include <array>
#include <condition_variable>
#include <mutex>

using ConstStringPair = std::pair<const std::string, const std::string>;
using HelpMessages = std::vector<ConstStringPair>;

class HostCOMDevice;
class VectorTable;

// 諸情報付きのアドレス
class saddr_t
{
 public:
	// コンストラクタ
	saddr_t() { }

	saddr_t(uint32 addr_, bool super_)
		: saddr_t(addr_, super_, false) { }

	saddr_t(uint32 addr_, bool super_, bool data_) {
		Set(addr_, super_, data_);
	}
	// コピーコンストラクタ
	saddr_t(const saddr_t& rhs) {
		Set(rhs);
	}
	// 代入演算子
	saddr_t& operator=(const saddr_t& rhs) {
		Set(rhs);
		return *this;
	}

	// 代入演算子 (uint32 を代入するとアドレスだけ差し替える)
	saddr_t& operator=(uint32 rhs) {
		SetAddr(rhs);
		return *this;
	}

	// 代入
	void Set(uint32 addr_, bool super_, bool data_) {
		addr  = addr_;
		super = super_;
		data  = data_;
	}

	uint32 GetAddr() const		{ return addr; }
	bool IsSuper() const		{ return super; }
	bool IsData()  const		{ return data; }
	void SetAddr(uint32 addr_)	{ addr  = addr_; }
	void SetSuper(bool super_)	{ super = super_; }
	void SetData(bool data_)	{ data  = data_; }

	// (uint32)saddr でアドレスだけ取り出せると楽
	explicit operator uint32() const { return GetAddr(); }

	// saddr += n でアドレス進められると楽
	uint32 operator+=(uint32 val) {
		addr += val;
		return addr;
	}
	// saddr++ でアドレス進められると楽
	void operator++(int n) {
		addr++;
	}

 private:
	void Set(const saddr_t& rhs) {
		addr  = rhs.addr;
		super = rhs.super;
		data  = rhs.data;
	}

	uint32 addr {};		// アドレス (32bit)
	bool super {};		// スーパーバイザアクセスなら true
	bool data  {};		// データ空間アクセスなら true
};

enum class MemoryMode : bool
{
	Logical  = false,
	Physical = true,
};

enum class MMULookupMode : bool
{
	False = false,
	True  = true,
	// MemoryMode::Physical ならこの項は意味を持たないので、
	// True でも False でもないと言いたい。
	None  = false,
};

// saddr_t の後で Debugger の前...
#include "debugger_memory.h"

class Debugger;
class Disasm;

// デバッガの CPU 依存部分
class DebuggerMD
{
 public:
	enum class Arch {
		M680x0,
		M88xx0,
	};

 public:
	DebuggerMD(Debugger *parent_, Arch arch_) {
		parent = parent_;
		arch = arch_;
	}
	virtual ~DebuggerMD();

	// CPU のリクエストフラグに flag を立てる
	virtual void ReqSet(uint32 flag) = 0;
	// CPU のリクエストフラグから flag を落とす
	virtual void ReqClr(uint32 flag) = 0;

	// 特権モードなら true を返す
	virtual bool IsSuper() const = 0;
	// MMU が有効なら true を返す
	virtual bool MMUEnabled() const = 0;

	// アドレス変換をする。
	// MMUEnabled() が true であることは呼び出し側が確認すること。
	// lookup が true ならテーブルサーチまで行う。
	// 変換できなければ (uint64)-1 を返す。
	virtual uint64 TranslateAddr(saddr_t laddr, bool lookup) const = 0;

	// 現在の命令先頭アドレスを返す
	virtual uint32 GetPC() const = 0;

	// ステップアウトを設定する
	virtual void SetStepOut() = 0;
	// ステップアウトなら true を返す
	virtual bool IsStepOut() const = 0;

	// この命令がステップイン出来るなら true を返す (cmd_n 用)
	virtual bool IsOpStepIn(DebuggerMemoryStream& mem) = 0;

	// name で示されるのレジスタの値を返す。
	// メモリダンプで指定するケースを想定しているので、明らかにアドレスでは
	// ないレジスタは含まなくてよい (m680x0 の SR など)。ただし Dn は含めて
	// よい。また m680x0 の SRP/CRP は下位32ビットでよい。
	// name が正しくない場合は (uint64)-1 を返す。
	virtual uint64 GetRegAddr(const char *name) const = 0;

	// 現在のレジスタセットを内部にバックアップする
	virtual void BackupRegs() = 0;

	// レジスタ表示系のコマンドを実行。処理すれば true を返す。
	virtual bool ShowRegister(FILE *, const std::vector<std::string>& args) = 0;
	// レジスタ表示系コマンドのヘルプ一覧を取得
	virtual const HelpMessages& GetHelpListReg() const = 0;
	// レジスタ表示系コマンドのヘルプメッセージを取得
	virtual const HelpMessages& GetHelpReg() const = 0;

	// ブランチ履歴・例外履歴(オブジェクト)を取得
	virtual BranchHistory& GetBrHist() const = 0;
	virtual BranchHistory& GetExHist() const = 0;

	// 直近の命令でアクセスした laddr が一致したら true を返す
	// 1 命令で複数にアクセスする場合を考慮して、md 側で判断する
	virtual bool CheckLEA(uint32 laddr) = 0;

	// 逆アセンブル。
	// mem.laddr の位置の命令を一つ逆アセンブルする。
	// 成功すれば true を返し、失敗すれば(?) false を返す。
	// 成功した場合 mnemonic に逆アセンブルした文字列が、
	// bin にこの命令のバイナリ列が格納されている。
	// mnemonic は単純に命令部分のみ、16進ダンプや付随情報は含まない。
	// bin はメモリイメージのバイト順で格納すること。
	virtual bool Disassemble(DebuggerMemoryStream& mem,
		std::string& mnemonic, std::vector<uint8>& bin) = 0;

	// 逆アセンブルをフォーマット。
	// オフライン版は逆アセンブルウィンドウなど、
	// オンライン版(Live) はデバッガプロンプトで表示するやつ。
	virtual std::string FormatDisasm(const std::string& mnemonic,
		const std::vector<uint8>& bin) = 0;
	virtual std::string FormatDisasmLive(const std::string& mnemonic,
		const std::vector<uint8>& bin) = 0;

	// 機種
	Arch arch {};

	// 最短の1命令のバイト数
	uint inst_bytes {};
	// 命令が固定長ならそのバイト数、可変長なら 0
	uint inst_bytes_fixed {};

 protected:
	Debugger *parent {};	// 親クラス
};

class Debugger : public ThreadDevice
{
	using inherited = ThreadDevice;
	friend class DebuggerMD_m680x0;
	friend class DebuggerMD_m88xx0;

 private:
	// リクエストフラグ
	static const uint32 REQUEST_EXIT	= 0x0001;	// 終了
	static const uint32 REQUEST_RXCHAR	= 0x0002;	// 1文字受信
	static const uint32 REQUEST_ACCEPT	= 0x0004;	// 着信通知
	static const uint32 REQUEST_PROMPT	= 0x0008;	// プロンプト要求

	// コマンド
	enum CmdAct {
		Stay,		// デバッガプロンプトに留まる
		Leave,		// デバッガプロンプトを抜けて VM を実行する
		Quit,		// デバッガとの接続を終了する
	};
	struct cmddef_t {
		const char *name;
		CmdAct (Debugger::*func)();
	};

	// ブレークポイント種別
	enum BreakpointType {
		Unused = 0,
		Address,		// このアドレスに来る直前
		Memory,			// このアドレスをアクセスした直後
		Exception,		// この例外を検出した直後
		Instruction,	// この命令を実行する直前
	};

	// ブレークポイント
	struct breakpoint_t {
		BreakpointType type {};	// 種別
		// 種別ごとのパラメータ
		union {
			uint32 addr {};		// ターゲットアドレス (Address/Memory)
			struct {
				int vec1;		// ベクタ番号 (開始)
				int vec2;		// ベクタ番号 (終了)
			};
			struct {
				uint32 inst;	// 命令ワード
				uint32 mask;	// マスク
			};
		};

		uint32 matched {};		// 成立回数 (積算)

		// skip はユーザ指定値。ただし
		// -1 なら常にスキップ、つまりブレークせずカウントのみ行う、
		// 0 ならスキップなし、つまり成立ごとにブレーク(これがデフォルト)、
		// n(>0) なら n 回の成立をスキップし、(n+1) 回目でブレークの意。
		int32 skip {};			// スキップ回数 (ユーザ指定値)
		int32 skipremain {};	// 残りスキップ回数 (0 で成立)
	};
	static const int MAX_BREAKPOINTS = 8;

 public:
	Debugger();
	~Debugger() override;

	void SetLogLevel(int) override;
	bool Create() override;
	bool Init() override;
	void ThreadRun() override;
	void Terminate() override;

	void Exec();
	void NotifyException(int vector);

	// MPU トレースが必要なら true を返す
	bool IsTrace() const;

	// FILE コールバック
	int ReadFunc(char *, int);
	int WriteFunc(const char *, int);

	// メモリダンプモニタ
	// GUI 用 (show コマンドでも見れるけど)
	std::array<Monitor, MAX_MEMDUMP_MONITOR> memdump_monitor {};

	// m/M コマンド用のメモリダンプモニタ
	Monitor m_monitor { this };

 private:
	void Close();
	void RxCallback();
	void AcceptCallback();
	void Input(int);
	void EnterPrompt();
	void LeavePrompt();
	void PrintPrompt();

	CmdAct Command();
	bool CheckAllBreakpoints();
	bool CheckBreakpointInst(breakpoint_t&);
	void ParseCmdbuf();
	// コマンド名は大文字小文字を区別する関係でスネークスタイル。
	CmdAct cmd_b();
	CmdAct cmd_bi();
	CmdAct cmd_bm();
	CmdAct cmd_bv();
	CmdAct cmd_b_list();
	CmdAct cmd_b_set(BreakpointType);
	CmdAct cmd_b_delete();
	CmdAct cmd_brhist();
	CmdAct cmd_bx();
	CmdAct cmd_c();
	CmdAct cmd_ct();
	CmdAct cmd_d();
	CmdAct cmd_dt();
	CmdAct cmd_D();
	CmdAct cmd_disp();
	CmdAct cmd_exhist();
	CmdAct cmd_h();
	CmdAct cmd_hb();
	CmdAct cmd_hr();
	CmdAct cmd_L();
	CmdAct cmd_m();
	CmdAct cmd_mt();
	CmdAct cmd_M();
	CmdAct cmd_minus();
	CmdAct cmd_n();
	CmdAct cmd_nt();
	CmdAct cmd_q();
	CmdAct cmd_reset();
	CmdAct cmd_s();
	CmdAct cmd_st();
	CmdAct cmd_so();
	CmdAct cmd_sot();
	CmdAct cmd_show();
	CmdAct cmd_t();

	bool Check();
	int AddBreakpoint(const breakpoint_t&);
	void RecalcInstMask();
	bool GetAddr(saddr_t *addr, bool);
	CmdAct cmd_d_common(MemoryMode access_mode, MMULookupMode lookup_mode);
	CmdAct cmd_hist_common(BranchHistory& hist);
	CmdAct cmd_m_common(MemoryMode access_mode, MMULookupMode lookup_mode);
	CmdAct cmd_n_common(bool);
	CmdAct cmd_s_common(bool);
	CmdAct cmd_so_common(bool);
	bool ParseVerbHex(const char *arg, uint32 *valp);
	bool ParseAddr(const char *arg, uint32_t *addrp);
	bool ParseTime(const char *arg, uint64 *timep);
	void SetNBreakpoint();

	static const HelpMessages HelpListMain;
	static const HelpMessages HelpListBreakpoints;
	static const HelpMessages HelpDetails;
	CmdAct ShowHelpList(const HelpMessages& msgs);

	// 個別ヘルプメッセージを出力用に置換
	std::string HelpConvert(const std::string& msg);

	// モニターを更新して表示
	void ShowMonitor(Monitor& monitor);
	// テキストスクリーンを表示
	void ShowTextScreen(TextScreen& screen);

	std::unique_ptr<DebuggerMD> md {};	// 機種依存部分

	std::mutex mtx {};
	// このスレッドへのリクエスト
	std::condition_variable cv_request {};
	uint32 request {};
	// コマンドモード(プロンプト)なら true
	std::condition_variable cv_prompt {};
	bool prompt_released {};

	bool is_prompt {};

	// コンソール
	std::unique_ptr<HostCOMDevice> hostcom /*{}*/;
	FILE *cons {};

	const char *prompt {};		// プロンプト文字列
	std::string cmdbuf {};		// 現在のコマンドライン
	std::string last_cmdbuf {};	// 直前のコマンドライン
	std::vector<std::string> args {};

	uint32 pc {};
	std::vector<uint8> ir {};	// この後実行する命令のバイナリ列
	std::array<breakpoint_t, MAX_BREAKPOINTS> bpoint {};
	bool bc_enable {};
	uint32 bc_addr {};
	uint32 bi_inst {};
	int bi_inst_bytes {};
	int bi_need_bytes {};
	int bv_vector = -1;
	bool ct_enable {};
	uint64 ct_end_time {};
	uint64 ct_timespan {};
	saddr_t d_last_addr {};
	saddr_t m_last_addr {};
	bool n_enable {};
	uint32 n_breakaddr {};
	uint32 n_count {};
	bool s_enable {};
	uint32 s_count {};
	bool so_enable {};
	bool t_enable {};
	uint32 t_count {};

	// 一度VMを実行して再びプロンプトに来たら true
	bool is_continued {};

	// MPU を一時停止するとき true (ワンショット)
	bool is_pause {};

	// プロンプトで表示するレジスタ群
	std::vector<std::string> disp_regs {};

	// ブレークポイント到達メッセージ
	// (コンソールが獲得できるまで保留しておくため)
	std::string bpointmsg {};

	// ブレークポイントモニタ
	DECLARE_MONITOR_CALLBACK(MonitorUpdateBpoint);
	Monitor bpoint_monitor { this };

	// メモリダンプモニタ
	DECLARE_MONITOR_CALLBACK(MonitorUpdateMemdump);

	// ベクタテーブル
	std::unique_ptr<VectorTable> pVectorTable /*{}*/;

	static std::vector<cmddef_t> cmdtable;
};

extern Debugger *gDebugger;

#define NORM		"\x1b[0m"
#define BOLD		"\x1b[1m"
#define BOLDIF(n)	((n) ? BOLD : "")
