//
// nono
// Copyright (C) 2022 nono project
// Licensed under nono-license.txt
//

//
// ベクタテーブル (モニタ)
//

#include "vectortable.h"
#include "bus.h"
#include "mainapp.h"
#include "mpu.h"

// グローバル参照用
VectorTable *gVectorTable;

// コンストラクタ
VectorTable::VectorTable(VMType vmtype)
	: inherited("(VectorTable)")
{
	// ログは不要
	ClearAlias();

	// 表示名を作成。それぞれ微妙に処理が違う…
	switch (vmtype) {
	 case VMType::X68030:
	 case VMType::LUNA1:
		InitTableM68030(vmtype);
		break;
	 case VMType::LUNA88K:
		InitTableLuna88k();
		break;
	 default:
		PANIC("invalid vmtype %d", (int)vmtype);
	}

	monitor.func = ToMonitorCallback(&VectorTable::MonitorUpdate);
	monitor.SetSize(49, 1 + 32);	// ヘッダの1行と全エントリ数
	monitor.SetMaxHeight(1 + Size());
	monitor.Regist(ID_SUBWIN_VECTOR);
}

// デストラクタ
VectorTable::~VectorTable()
{
	gVectorTable = NULL;
}

// X68030、LUNA-I の表示名テーブル初期化
void
VectorTable::InitTableM68030(VMType vmtype)
{
	nametable.clear();
	nametable.resize(256);

	// まず m68k 標準の名称で埋める
	for (int v = 0; v < 64; v++) {
		if (name_m68030[v]) {
			nametable[v] = name_m68030[v];
		}
	}

	if (vmtype == VMType::X68030) {
		// X68030 なら、
		// 全域で、あれば X68030 用の名前で上書き
		for (int v = 0; v < nametable.size(); v++) {
			if (name_x68030[v]) {
				nametable[v] = name_x68030[v];
			}
		}
	} else {
		// LUNA-I なら、
		// レベル割り込み名だけ上書き
		for (int i = 0; i < countof(name_luna1); i++) {
			nametable[0x19 + i] = name_luna1[i];
		}
	}
}

// LUNA-88K の表示名テーブルを初期化
void
VectorTable::InitTableLuna88k()
{
	nametable.clear();
	nametable.resize(512);

	// 先頭から10個くらいだけ連続しているのでテーブルからコピー
	for (int i = 0; i < countof(name_luna88k); i++) {
		nametable[i] = name_luna88k[i];
	}

	// 以降は不連続で、間が空きすぎているので、ここで代入

	// SFU
	// (SFU2〜7 はいいか…)
	//                01234567890123456789012 <- 例外履歴欄の横幅
	nametable[114] = "SFU1 Precise Exception";
	nametable[115] = "SFU1 ImpreciseException";

	// OpenBSD
	// XXX 本来は OpenBSD 稼働時に限定すべきだろうけど、そうする意味もない
	//                01234567890123456789012
	nametable[450] = "OpenBSD System Call";
	nametable[451] = "OpenBSD Cache Flush";
	nametable[503] = "Division by Zero(GCC)";
	nametable[504] = "OpenBSD stepbpt";
	nametable[511] = "OpenBSD userbpt";
}

// 例外名を返す (ベクタテーブル用)。
// 名前がなければ "" を返す。
const char *
VectorTable::GetTableName(int vector) const
{
	assertmsg(0 <= vector && vector < nametable.size(), "vector=%d", vector);

	return nametable[vector] ?: "";
}

// 例外名を返す (例外履歴用)。
// 名前がなければ NULL を返す。
const char *
VectorTable::GetExceptionName(int vector) const
{
	assertmsg(0 <= vector && vector < nametable.size(), "vector=%d", vector);

	if (__predict_false(vector < 2)) {
		if (gMainApp.Has(VMCap::M68K)) {
			// ベクタテーブル的には 0:"Reset(SP)"、1:"Reset(PC)" のほうが
			// 分かりやすいが、例外の名前は 0 が "Reset Exception" で
			// 1 は存在しない。
			if (vector == 0) {
				return "Reset Exception";
			} else {
				return "(vector 1?)";
			}
		}
	}

	if (__predict_true(nametable[vector] != NULL)) {
		return nametable[vector];
	} else {
		return NULL;
	}
}

// モニタ更新
void
VectorTable::MonitorUpdate(Monitor *, TextScreen& screen)
{
	// 0         1         2         3         4         5         6
	// 0123456789012345678901234567890123456789012345678901234567890123456789
	// No.      Offset Address   Name                         Count
	// $02(  2) $0008  $12345678 01234567890123456789012 0123456789

	uint32 vbr = gMPU->GetVBR();

	// userdata が表示開始位置
	int v = (int)screen.userdata;
	int row = screen.GetRow();
	int vend = v + row - 1;

	screen.Clear();

	// 最初の1行は常にヘッダ
	screen.Print(0, 0, "No.      Offset Address   Name");
	//screen.Print(57, 0, "Count");

	for (int y = 1; v < vend; v++, y++) {
		screen.Print(0, y, "$%02x(%3d) $%04x", v, v, v * 4);
		uint64 addr = vm_phys_peek_32(vbr + v * 4);
		if (__predict_false((int64)addr < 0)) {
			screen.Print(16, y, "BusErr");
		} else {
			screen.Print(16, y, "$%08x", (uint32)addr);
		}
		screen.Print(26, y, "%-23s", GetTableName(v));
	}
}

// ベクタ名 (MC68030 共通)
/*static*/ std::array<const char * const, 64> VectorTable::name_m68030 = {
			//   0123456789012345678 <- 例外履歴欄の横幅
	/*  0 */	"Reset (SP)",
	/*  1 */	"Reset (PC)",
	/*  2 */	"Bus Error",
	/*  3 */	"Address Error",
	/*  4 */	"Illegal Instruction",
	/*  5 */	"Zero Divide",
	/*  6 */	"CHK/CHK2 Insn",
	/*  7 */	"TRAPV, *TRAPcc Insn",
	/*  8 */	"PriviledgeViolation",
	/*  9 */	"Trace",
	/* 10 */	"Line 1010 emulator",
	/* 11 */	"Line 1111 emulator",
	/* 12 */	NULL,
	/* 13 */	"CoproProtoViolation",
	/* 14 */	"Format Error",
	/* 15 */	"Uninitialized Intr",
	/* 16 */	NULL, NULL, NULL, NULL,
	/* 20 */	NULL, NULL, NULL, NULL,
	/* 24 */	"Spurious Interrupt",
	/* 25 */	"Level 1 Interrupt",
	/* 26 */	"Level 2 Interrupt",
	/* 27 */	"Level 3 Interrupt",
	/* 28 */	"Level 4 Interrupt",
	/* 29 */	"Level 5 Interrupt",
	/* 30 */	"Level 6 Interrupt",
	/* 31 */	"Level 7 Interrupt",
	/* 32 */	"Trap #0",
	/* 33 */	"Trap #1",
	/* 34 */	"Trap #2",
	/* 35 */	"Trap #3",
	/* 36 */	"Trap #4",
	/* 37 */	"Trap #5",
	/* 38 */	"Trap #6",
	/* 39 */	"Trap #7",
	/* 40 */	"Trap #8",
	/* 41 */	"Trap #9",
	/* 42 */	"Trap #10",
	/* 43 */	"Trap #11",
	/* 44 */	"Trap #12",
	/* 45 */	"Trap #13",
	/* 46 */	"Trap #14",
	/* 47 */	"Trap #15",
			//   0123456789012345678
	/* 48 */	"FPCP Branch",
	/* 49 */	"FPCP InexcactResult",
	/* 50 */	"FPCP Divide by Zero",
	/* 51 */	"FPCP UnderFlow",
	/* 52 */	"FPCP Operand Error",
	/* 53 */	"FPCP OverFlow",
	/* 54 */	"FPCP Signaling NAN",
	/* 55 */	NULL,
	/* 56 */	"MMU Config Error",
	/* 57 */	NULL,
	/* 58 */	NULL,
	/* 59 */	NULL,
	/* 60 */	NULL,
	/* 61 */	NULL,
	/* 62 */	NULL,
	/* 63 */	NULL,
};

// ベクタ名 (X68030)
/*static*/ std::array<const char * const, 256> VectorTable::name_x68030 = {
	/* $00 */	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	/* $08 */	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	/* $10 */	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	/* $18 */	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	/* $20 */	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	/* $28 */	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	/* $30 */	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	/* $38 */	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,

			//   0123456789012345678
	/* $40 */	"MFP Alarm",
	/* $41 */	"MFP EXPON",
	/* $42 */	"MFP POWSW",
	/* $43 */	"MFP FM IRQ",
	/* $44 */	"MFP Timer-D",
	/* $45 */	"MFP Timer-C",
	/* $46 */	"MFP V-Disp",
	/* $47 */	"MFP GPIP5",
	/* $48 */	"MFP Timer-B",
	/* $49 */	"MFP TX Error",
	/* $4a */	"MFP TX Empty",
	/* $4b */	"MFP RX Error",
	/* $4c */	"MFP RX Full",
	/* $4d */	"MFP Timer-A",
	/* $4e */	"MFP CRTC IRQ",
	/* $4f */	"MFP H-Sync",
	/* $50 */	"SCC#B TX Empty",
	/* $51 */	NULL,
	/* $52 */	"SCC#B E/S",
	/* $53 */	NULL,
	/* $54 */	"SCC#B RX Intr",
	/* $55 */	NULL,
	/* $56 */	"SCC#B SpCond",
	/* $57 */	NULL,
	/* $58 */	"SCC#A TX Empty",
	/* $59 */	NULL,
	/* $5a */	"SCC#A E/S",
	/* $5b */	NULL,
	/* $5c */	"SCC#A RX Intr",
	/* $5d */	NULL,
	/* $5e */	"SCC#A SpCond",
	/* $5f */	NULL,

			//   0123456789012345678
	/* $60 */	"I/O FDC Intr",
	/* $61 */	"I/O FDD Intr",
	/* $62 */	"I/O HDC Intr",
	/* $63 */	"I/O PRN Intr",
	/* $64 */	"DMAC#0 Complete",
	/* $65 */	"DMAC#0 Error",
	/* $66 */	"DMAC#1 Complete",
	/* $67 */	"DMAC#1 Error",
	/* $68 */	"DMAC#2 Complete",
	/* $69 */	"DMAC#2 Error",
	/* $6a */	"DMAC#3 Complete",
	/* $6b */	"DMAC#3 Error",
	/* $6c */	"SPC Interrupt",
	/* $6d */	NULL,
	/* $6e */	NULL,
	/* $6f */	NULL,

	/* $70 */	NULL, NULL, NULL, NULL,
	/* $74 */	NULL, NULL, NULL, NULL,
	/* $78 */	NULL, NULL, NULL, NULL,
	/* $7c */	NULL, NULL, NULL, NULL,
	/* $80 */	NULL, NULL, NULL, NULL,	// MIDI
	/* $84 */	NULL, NULL, NULL, NULL,	// MIDI
	/* $88 */	NULL, NULL, NULL, NULL,	// MIDI
	/* $8c */	NULL, NULL, NULL, NULL,	// MIDI
	/* $90 */	NULL, NULL, NULL, NULL,
	/* $94 */	NULL, NULL, NULL, NULL,
	/* $98 */	NULL, NULL, NULL, NULL,
	/* $9c */	NULL, NULL, NULL, NULL,
	/* $a0 */	NULL, NULL, NULL, NULL,	// MIDI
	/* $a4 */	NULL, NULL, NULL, NULL,	// MIDI
	/* $a8 */	NULL, NULL, NULL, NULL,	// MIDI
	/* $ac */	NULL, NULL, NULL, NULL,	// MIDI
	/* $b0 */	NULL, NULL, NULL, NULL,	// RS-232C
	/* $b4 */	NULL, NULL, NULL, NULL,	// RS-232C
	/* $b8 */	NULL, NULL, NULL, NULL,	// RS-232C
	/* $bc */	NULL, NULL, NULL, NULL,	// RS-232C
	/* $c0 */	NULL, NULL, NULL, NULL,	// RS-232C
	/* $c4 */	NULL, NULL, NULL, NULL,	// RS-232C
	/* $c8 */	NULL, NULL, NULL, NULL,	// RS-232C
	/* $cc */	NULL, NULL, NULL, NULL,	// RS-232C
	/* $d0 */	NULL, NULL, NULL, NULL,
	/* $d4 */	NULL, NULL, NULL, NULL,
	/* $d8 */	NULL, NULL, NULL, NULL,
	/* $dc */	NULL, NULL, NULL, NULL,
	/* $e0 */	NULL, NULL, NULL, NULL,
	/* $e4 */	NULL, NULL, NULL, NULL,
	/* $e8 */	NULL, NULL, NULL, NULL,
	/* $ec */	NULL, NULL, NULL, NULL,

			//   0123456789012345678
	/* $f0 */	"PSX16x50 Interrupt",
	/* $f1 */	NULL,
	/* $f2 */	NULL,
	/* $f3 */	NULL,
	/* $f4 */	NULL,
	/* $f5 */	NULL,
	/* $f6 */	"ExSPC Interrupt",
	/* $f7 */	NULL,	// TS-6BSI
	/* $f8 */	"Nereid#1 Interrupt",
	/* $f9 */	"Neptune-X Interrupt",
	/* $fa */	"Nereid#1 USB",
	/* $fb */	"Nereid#0 USB",
	/* $fc */	NULL,
	/* $fd */	NULL,
	/* $fe */	NULL,
	/* $ff */	NULL,
};

// LUNA-I はレベル割り込みのみ
/*static*/ std::array<const char * const, 7> VectorTable::name_luna1 = {
	//      	 0123456789012345678 <- 例外履歴欄の横幅
	/* $19 */	"Lv1 Intr (XP Low)",
	/* $1a */	"Lv2 Intr (SPC)",
	/* $1b */	"Lv3 Intr (Lance)",
	/* $1c */	"Lv4 Intr (XP High)",
	/* $1d */	"Lv5 Intr (SysClk)",
	/* $1e */	"Lv6 Intr (SIO)",
	/* $1f */	"Lv7 Intr (NMI)",
};

// LUNA-88K は先頭11個のみ (残りはコードで処理)
/*static*/ std::array<const char * const, 11> VectorTable::name_luna88k = {
			//   01234567890123456789012 <- 例外履歴欄の横幅
	/*  0 */	"Reset Exception",
	/*  1 */	"Interrupt Exception",
	/*  2 */	"Inst Access Exception",
	/*  3 */	"Data Access Exception",
	/*  4 */	"Misaligned Access Excep",
	/*  5 */	"Unimplemented Opcode",
	/*  6 */	"Priv. Violation Excep.",
	/*  7 */	"Bounds Check Violation",
	/*  8 */	"Illegal Integer Divide",
	/*  9 */	"Int Overflow Exception",
	/* 10 */	"Error Exception",
			//   01234567890123456789012
};
