//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// 設定
//

#include "config.h"
#include "fdc.h"
#include "mystring.h"

// グローバル参照用
Config *gConfig;

// コンストラクタ
Config::Config()
{
	// 設定項目はすべてここで用意すること。
	// ここで追加した順に表示されるので基本的にはアルファベット順に並べる
	// (先頭のドットは考慮しなくていい)。
	// ただし vmtype だけは特別なので最初に表示されたほうがいいだろう。

	Add("vmtype", "");

	// 以下概ねアルファベット順に並べる

	// X68030 CGROM ファイルパス。なければ内蔵 CGROM。
	Add("cgrom-image", "", VMCap::X68030);

	// VM 内の時計を同期する時間軸。
	// [ real, virtual ]
	Add("clock-sync", "real");

	// デバッガコンソール
	Add("debugger-driver", "none");
	Add("debugger-tcp-port", 0);
	// デバッガコンソールの fallback は常に不可とする。hostcom.cpp 参照。
	//  "debugger-fallback"

	// MAC アドレス
	Add("ethernet-macaddr", "auto", VMCap::LUNA);

	// フロッピーディスク
	Add("fd-drive", 2, VMCap::X68030);
	for (int id = 0; id < FDCDevice::MAX_DRIVE; id++) {
		// 種別とイメージパス
		Add(string_format("fd%d-image", id), VMCap::X68030);
		// エミュレータ的に書き込みを禁止するか
		Add(string_format("fd%d-writeignore", id), 0, VMCap::X68030);
	}

	// X68030 時の FPU の有無。
	Add("fpu-type", "68881", VMCap::X68030);

	// シリアルポートのホスト側設定
	Add("hostcom-driver", "none");
	Add("hostcom-fallback", 0);
	Add("hostcom-tcp-port", 9998);

	// キーボードの入力モード
	Add("hostkbd-input", "char");

	// ネットワークのホスト側設定。
	Add("hostnet-driver", "auto");
	Add("hostnet-fallback", 0);
	Add("hostnet-afpacket-ifname", "auto");
	Add("hostnet-bpf-ifname", "auto");
	Add("hostnet-tap-devpath", "auto");

	// IPLROM1/2 ファイルパス。なければ内蔵 ROM。
	Add("iplrom1-image", "", VMCap::X68030);
	Add("iplrom2-image", "", VMCap::X68030);

	// キーボードの接続
	Add("keyboard-connect", 1);

	// LUNA の RTC(MK48T02) がうるう年でない 1970 をエポックとして
	// 設定しているので、そちらに合わせてうるう年を補正する
	Add("luna-adjust-misused-epoch", 1, VMCap::LUNA);

	// LUNA の DIPSW (1 が UP、0 が DOWN)。
	// 初期値は機種によって異なるため pio.cpp で設定している。
	Add("luna-dipsw1", VMCap::LUNA);
	Add("luna-dipsw2", VMCap::LUNA);

	// LUNA ビデオボードプレーン数
	Add("luna-video-plane", 1, VMCap::LUNA);

	// m88100 の別名ニーモニックを併記する。
	Add(".m88k-altname", 0, VMCap::M88K);

	// モニタウィンドウの更新頻度 [Hz]
	Add("monitor-rate", 20);

	// MPU クロック (MHz 単位の実数、小数点以下は3桁まで)。
	// 初期値は各 VM クラスで設定している。
	Add("mpu-clock");

	// M88100 で疑似 STOP 状態をサポートする
	Add("mpu-pseudo-stop", 1, VMCap::M88K);

	// PROM ファイルパス。なければ内蔵 ROM 起動。
	Add("prom-image", "", VMCap::LUNA);

	// PROM のデバッグオプション。
	// PROM 中の SCSI デバッグビットを立てる。
	Add(".prom-scsi-debug", 0, VMCap::LUNA1);

	// RAM サイズ (MB 単位の整数)。
	// 初期値は機種ごとの VM クラスで設定している。
	Add("ram-size", 0);

	// RTC のデバッグ用オプション。
	Add(".rtc-force-fixed", 0);
	Add(".rtc-date");
	Add(".rtc-time");

	// RTC の epoch。初期値は各 VM クラスで設定している。
	Add("rtc-epoch-year", VMCap::LUNA);

	// ステータスパネルを表示する。
	Add("show-statuspanel", 1);

	// SCSI 設定。
	for (int id = 0; id < 8; id++) {
		// 種別とイメージパス
		Add(string_format("spc0-id%d-image", id));
		// エミュレータ的に書き込みを禁止するか
		Add(string_format("spc0-id%d-writeignore", id), 0);
		// エミュレータ的に書き込みを禁止するか
		// (後方互換、十分不要になった頃に削除する)
		Add(string_format("spc0-id%d-writeprotect", id), 0);
		// シークタイム [msec]
		Add(string_format("spc0-id%d-seektime", id), 0);
	}

	// SRAM の RAM 容量を同期するか。
	Add("sram-sync-ramsize", 1, VMCap::X68030);
}

// デストラクタ
Config::~Config()
{
	gConfig = NULL;
}

// 項目を追加 (初期状態用)。値は空文字列。
bool
Config::Add(const std::string& key, VMCap vmcap)
{
	return Add(key, "", vmcap);
}

// 項目を追加 (初期状態用)。値は整数値を文字列にしたもの。
bool
Config::Add(const std::string& key, int value, VMCap vmcap)
{
	return Add(key, std::to_string(value), vmcap);
}

// 項目を追加 (初期状態用)
bool
Config::Add(const std::string& key, const std::string& value, VMCap vmcap)
{
	assert(FindItem(key) == NULL);
	assert(fixed == false);

	list.emplace_back(key, value, vmcap, ConfigItem::FromInitial, "");
	return true;
}

// 項目を削除
// key が見付からなくても何もしない。key は複数登録されていることはないはず。
void
Config::Delete(const std::string& key)
{
	for (auto it = list.begin(); it != list.end(); ++it) {
		auto item = *it;
		if (key == item.key) {
			list.erase(it);
			return;
		}
	}
	assert(FindItem(key) == NULL);
}

// 設定ファイルの内容で上書きする
bool
Config::Update(const ConfigFile& file)
{
	assert(fixed == false);

	for (const auto& pair : file.lines) {
		const std::string& line = pair.first;
		int lineno = pair.second;

		auto where = string_format("%s:%d", file.filepath.c_str(), lineno);
		if (UpdateLine(line, ConfigItem::FromConfig, where) == false) {
			return false;
		}
	}

	return true;
}

// コマンドラインオプションで更新
bool
Config::Update(const std::vector<std::string>& options)
{
	assert(fixed == false);

	for (const auto& line : options) {
		if (UpdateLine(line, ConfigItem::FromOption, "") == false) {
			return false;
		}
	}

	return true;
}

// 更新1行分の共通処理。
// キーが見付からない場合、コマンドラインからならエラーとし、
// それ以外からの場合は警告のみで無視する。
// エラーが起きたらエラーメッセージを表示して false を返す。
bool
Config::UpdateLine(const std::string& line, ConfigItem::from_t from,
	const std::string& where)
{
	std::string key;
	std::string val;

	// キーと値に分解
	if (SplitKeyVal(&key, &val, line) == false) {
		ConfigItem::PrintErr(from, where, "", line, "syntax error");
		return false;
	}

	// キーで検索
	ConfigItem *itemptr = FindItem(key);
	if (itemptr == NULL) {
		// コマンドラインからの場合はエラーにする
		if (from == ConfigItem::FromOption) {
			ConfigItem::PrintErr(from, where, key, val, "unknown key");
			return false;
		}
		// コマンドライン以外の場合は警告のみで無視する
		ConfigItem::PrintErr(from, where, key, val, "unknown key, ignored");
		return true;
	}

	// 見付かったので更新
	ConfigItem& item = *itemptr;
	item.value = val;
	item.from  = from;
	item.where = where;

	return true;
}

// "key=val" 形式の行 line を分解するして *keyp, *valp に代入する。
// その際 key, val 前後の空白は取り除く。
// 失敗すれば false を返す。
/*static*/ bool
Config::SplitKeyVal(std::string *keyp, std::string *valp,
	const std::string& line)
{
	assert(keyp);
	assert(valp);

	// キーと値に分解
	auto pos = line.find('=');
	if (pos == std::string::npos) {
		return false;
	}
	auto key = line.substr(0, pos++);
	auto val = line.substr(pos, line.size() - pos);

	*keyp = string_trim(key);
	*valp = string_trim(val);

	return true;
}

// 内容を表示する。
// all なら隠し設定も含めて表示。
void
Config::Show(bool all) const
{
	assert(fixed);

	for (const auto& item : list) {
		// '!' から始まるのは内部用変数なので隠す
		if (!all && item.key[0] == '!') {
			continue;
		}
		// ドットから始まるのは隠し設定で、初期値のままなら隠す
		if (!all &&
		    item.key[0] == '.' && item.from == ConfigItem::FromInitial) {
			continue;
		}

		std::string buf = string_format("%s=\"%s\"",
			item.key.c_str(), item.value.c_str());
		if (buf.size() < 40) {
			int n = ((40 - buf.size()) + 7) / 8;
			buf.append(n, '\t');
		} else {
			buf.push_back('\t');
		}

		std::string fromstr;
		switch (item.from) {
		 case ConfigItem::FromInitial:
			fromstr = "initial value";
			break;
		 case ConfigItem::FromConfig:
			fromstr = item.where;
			break;
		 case ConfigItem::FromOption:
			fromstr = "-V option";
			break;
		 default:
			__unreachable();
		}

		printf("%s(%s)\n", buf.c_str(), fromstr.c_str());
	}
}

// key から ConfigItem を取得する。見付からなければ assert する。
const ConfigItem&
Config::Find(const std::string& key) const
{
	for (const auto& item : list) {
		if (key == item.key) {
			return item;
		}
	}
	PANIC("Config::Find() \"%s\" not found", key.c_str());
}

// key のデフォルト値を差し替える。key が見付からなければ assert する。
void
Config::SetDefault(const std::string& key, int val)
{
	std::string valstr = std::to_string(val);
	SetDefault(key, valstr);
}

// key のデフォルト値を差し替える。key が見付からなければ assert する。
void
Config::SetDefault(const std::string& key, const std::string& val)
{
	assert(fixed == false);

	for (auto& item : list) {
		if (key == item.key) {
			item.SetDefault(val);
			return;
		}
	}
	assertmsg(0, "Config::SetDefault() \"%s\" not found\n", key.c_str());
}

// key から ConfigItem を返す。見付からなければ NULL を返す。
// (こっちは内部で変更するために使う)
ConfigItem *
Config::FindItem(const std::string& key)
{
	for (auto& item : list) {
		if (key == item.key) {
			return &item;
		}
	}
	return NULL;
}

// これ以降は書き込み禁止
void
Config::Fix()
{
	assert(fixed == false);

	// ここで、この機種に不要な変数を削除する。
	for (auto it = list.begin(); it != list.end(); ) {
		ConfigItem& item = *it;
		VMCap vmcap = item.GetVMCap();
		if (gMainApp.Has(vmcap) == false) {
			it = list.erase(it);
		} else {
			++it;
		}
	}

	fixed = true;
}


//
// 設定項目
//

// 値を整数値として取得してみる。
bool
ConfigItem::TryInt(int *val) const
{
	char *endp;

	if (value.empty()) {
		return false;
	}

	errno = 0;
	int r = strtol(value.c_str(), &endp, 10);
	if (errno == ERANGE || *endp != '\0') {
		return false;
	}
	*val = r;
	return true;
}

// デフォルト値を差し替える
void
ConfigItem::SetDefault(const std::string& val)
{
	if (from == FromInitial) {
		value = val;
	}
}

void
ConfigItem::Err() const
{
	// 定型文
	Err("invalid argument");
}

void
ConfigItem::Err(const char *fmt, ...) const
{
	char buf[1024];
	va_list ap;

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);

	std::string msg(buf);
	Err(msg);
}

void
ConfigItem::Err(const std::string& msg) const
{
	PrintErr(from, where, key, value, msg);
}

/*static*/ void
ConfigItem::PrintErr(ConfigItem::from_t from, const std::string& where,
	const std::string& key, const std::string& value, const std::string& msg)
{
	std::string fromstr;
	std::string keyvalue;

	switch (from) {
	 case FromInitial:
		fromstr = "initial value:";
		break;
	 case FromConfig:
		fromstr = where + ":";
		break;
	 case FromOption:
		fromstr = "option -V";
		break;
	 default:
		__unreachable();
	}

	if (key.empty()) {
		keyvalue = value;
	} else {
		keyvalue = key + "=" + value;
	}

	warnx("%s %s: %s", fromstr.c_str(), keyvalue.c_str(), msg.c_str());
}


//
// 設定ファイル
//

// コンストラクタ
ConfigFile::ConfigFile(const std::string& filepath_)
{
	filepath = filepath_;
}

// デストラクタ
ConfigFile::~ConfigFile()
{
}

// ファイルの読み込み
bool
ConfigFile::Load()
{
	char buf[1024];
	FILE *fp;

	fp = fopen(filepath.c_str(), "r");
	if (fp == NULL) {
		warn("%s", filepath.c_str());
		return false;
	}

	for (int line = 1; fgets(buf, sizeof(buf), fp) != NULL; line++) {
		// 空行とコメント行を無視
		rtrim(buf);
		if (buf[0] == '#') {
			buf[0] = '\0';
		}
		if (buf[0] == '\0') {
			continue;
		}

		// ここでは記憶するだけ
		lines.emplace_back(buf, line);
	}

	fclose(fp);
	return true;
}
