//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// 基本ヘッダ
//

#pragma once

#define NONO_MAJOR_VER	(0)
#define NONO_MINOR_VER	(4)
#define NONO_PATCH_VER	(1)
#define NONO_DATE		"2022/07/24"

#include "config-nono.h"
#include <cassert>
#include <cctype>
#include <cerrno>
#include <cstdarg>
#include <cstdio>
#include <cstdint>
#include <cstdlib>
#include <cstring>
#include <memory>
#include <string>

#include <err.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/cdefs.h>
#include <sys/types.h>

using uint8		= uint8_t;
using uint16	= uint16_t;
using uint32	= uint32_t;
using uint64	= uint64_t;
using int8		= int8_t;
using int16		= int16_t;
using int32		= int32_t;
using int64		= int64_t;

#if defined(HAVE_BSD_BSD_H)
#include <bsd/bsd.h>
#endif

#include "myendian.h"

#if !defined(__printflike)
#if defined(HAVE___ATTRIBUTE_FORMAT)
# define __printflike(a,b)	__attribute__((__format__(__printf__, (a), (b))))
#else
# define __printflike(a,b)
#endif
#endif

#if !defined(__unreachable)
#if defined(HAVE___BUILTIN_UNREACHABLE)
# define __unreachable()	__builtin_unreachable()
#else
# define __unreachable()	((void)0)
#endif
#endif

// gcc でコンパイラに最適化の条件を与える。
#if defined(HAVE___BUILTIN_ASSUME)
#define __assume(cond)	__builtin_assume(cond)
#else
#define __assume(cond)	if (!(cond)) __unreachable()
#endif

#if !defined(__packed)
#if defined(HAVE___ATTRIBUTE_PACKED)
# define __packed	__attribute__((__packed__))
#else
# error not tested.
#endif
#endif

// FreeBSD(12.1) の __predict_true/false は定義がいまいちで使えない。
// __builtin_expect() があることが分かればどの環境でも自前で定義できるので
// 既存定義は取り消す。
#if defined(HAVE___BUILTIN_EXPECT)
# undef  __predict_true
# undef  __predict_false
# define __predict_true(exp)	__builtin_expect((exp) != 0, 1)
# define __predict_false(exp)	__builtin_expect((exp) != 0, 0)
#elif !defined(__predict_true)
# define __predict_true(exp)	(exp)
# define __predict_false(exp)	(exp)
#endif

#if !defined(__unused)
#if defined(HAVE___ATTRIBUTE_UNUSED)
# define __unused __attribute__((__unused__))
#else
# define __unused
#endif
#endif

#if defined(__clang__)
# define FALLTHROUGH	[[clang::fallthrough]]
#elif defined(HAVE___ATTRIBUTE_FALLTHROUGH)
# define FALLTHROUGH	__attribute__((__fallthrough__))
#else
# define FALLTHROUGH	((void)0)
#endif

#if defined(HAVE___ATTRIBUTE_NO_SANITIZE)
# define NO_SANITIZE(x)	__attribute__((no_sanitize(x)))
#else
# define NO_SANITIZE(x)
#endif

// x86_64 では SSE のために16バイト境界に整列させたいが、ほかの
// アーキテクチャでは alignas(16) がエラーになる場合がある。この場合の
// alignas() は SSE のために必要なので、他アーキテクチャではそもそも
// 指定しなくていい。ということで x86_64 の時に限って指定したい時に使う。
// うーん、なんだこれ。
#if defined(__x86_64__)
#define ALIGNAS_IF_X86(n) alignas(n)
#else
#define ALIGNAS_IF_X86(n) /**/
#endif

// 64ビット数と2つの32ビット数を扱う共用体。
// 構造はまったく同じだが名前だけ場所に合わせてつけたい。
#if BYTE_ORDER == LITTLE_ENDIAN
#define DEF_UNION64(NAME, Q, H, L)	\
	union NAME {	\
		uint64 Q;	\
		struct { uint32 L, H; };	\
	}
#else
#define DEF_UNION64(NAME, Q, H, L)	\
	union NAME {	\
		uint64 Q;	\
		struct { uint32 H, L; };	\
	}
#endif

DEF_UNION64(union64, q, h, l);

#ifndef countof
#define countof(x)	(sizeof(x) / sizeof(x[0]))
#endif

#ifndef roundup
#define roundup(x, y)	((((x)+((y)-1))/(y))*(y))
#endif

#ifndef rounddown
#define rounddown(x, y)	(((x)/(y))*(y))
#endif

// メッセージ付きの assert。
// 本当は後ろに足したいのだがそれは無理なので、せめてそれっぽく1行にしてみる。
#if !defined(NDEBUG)
#define assertmsg(expr, ...)	do {	\
	if (__predict_false(!(expr))) {	\
		fprintf(stderr, __VA_ARGS__);	\
		fprintf(stderr, ": "); \
		assert(expr);	\
	}	\
} while (0)
#else
#define assertmsg(expr, ...)
#endif

// atomic_compare_exchange_strong() の短縮マクロ。
// ライブラリ実装は第2引数の expected が値ではなくポインタなので一旦代入操作を
// 行わないといけないのと、第3引数の newval は型が uint8 だと int 即値が書けず
// (int8) でキャストしないといけない。
// その上 atomic_compare_exchange_strong() が長い、と、いろいろ微妙なので、
// その辺を短くするマクロを用意する。ただしこれでは expected 変数の型が
// 自動判別できないので、_8 と _32 を用意するところは妥協。
// (static) inline 関数にすることも出来るけどそうするとここで <atomic> ヘッダ
// を読み込まなければならなくなるため、マクロで実現する。
#define atomic_cas_8(varp, expval, newval)	({	\
	uint8 expected = (expval);	\
	atomic_compare_exchange_strong((varp), &expected, (uint8)(newval));	\
})
#define atomic_cas_32(varp, expval, newval)	({	\
	uint32 expected = (expval);	\
	atomic_compare_exchange_strong((varp), &expected, (uint32)(newval));	\
})

// パニック (PC を表示しない)
#define PANIC(...)	panic(__PRETTY_FUNCTION__, __VA_ARGS__)
[[noreturn]] extern void panic(const char *funcname, const char *fmt, ...)
	__printflike(2, 3);

// VM 用パニック (PC を表示する)
#define VMPANIC(...) vmpanic(__PRETTY_FUNCTION__, __VA_ARGS__)
[[noreturn]] extern void vmpanic(const char *funcname, const char *fmt, ...)
	__printflike(2, 3);
