//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// CLI/GUI 共通のメイン部分
//

#pragma once

#include "object.h"

class Config;
class Logger;
class MonitorManager;
class VM;

// VM 種別
enum class VMType {
	NONE = 0,
	X68030,
	LUNA1,
	LUNA88K,
};

// VM ケーパビリティ (ビットマスク表現)
enum class VMCap : uint32 {
	NONE	= 0,
	X68030	= (1U << (int)VMType::X68030),
	LUNA1	= (1U << (int)VMType::LUNA1),
	LUNA88K	= (1U << (int)VMType::LUNA88K),

	LUNA	= VMCap::LUNA1 | VMCap::LUNA88K,
	M68K	= VMCap::X68030 | VMCap::LUNA1,
	M88K	= VMCap::LUNA88K,
	ALL		= 0xffffffff,
};

class MainApp
{
 public:
	// Init() の戻り値
	static const int PASS = -1;		// 実行を継続

 public:
	MainApp();
	~MainApp();

	// 初期化 (ステージ1)
	int Init1(bool is_cli_, int ac, char *av[]);

	// 初期化 (ステージ2)
	bool Init2();

	std::string SearchFile(const std::string& name) const;

	// VM 種別を返す。
	// (Config が持っているのは文字列)
	VMType GetVMType() const { return vmtype; }

	// 現在の VM が指定のケーパビリティを持っているか?
	bool Has(VMCap cap) const;

	// VM ディレクトリを返す
	const std::string& GetVMDir() const { return vmdir; }

	// ログレベルを設定する。
	static bool SetLogopt(const std::vector<std::string>&, std::string *errmsg);

	// ログ名の一覧を表示する
	static std::vector<std::string> GetLogNames();

	// オブジェクトを追加/削除 (Object が呼ぶ)
	void AddObject(Object *);
	void DeleteObject(Object *);

	// オブジェクトのリストを返す
	const std::vector<Object *>& GetObjects() const { return objects; }

	// Logger を返す
	Logger *GetLogger() const { return logger.get(); }

 public:
	// -B: ベンチマークモード
	int benchmark_mode {};

	// -b: ブレークポイント "<addr>[,<skip>]"
	std::vector<std::string> debug_breakaddr {};

	// -d: 起動時にデバッガで停止
	bool debug_on_start {};

	// -D: デバッガを stdio で使う
	bool opt_D {};

	// -f: 高速モード (XXX 設定ファイルと含めて見直すこと)
	bool fast_mode {};

	// --fontsize: フォントサイズ指定値 (値は fontsize_t ではない)
	int  fontsize {};

	// -H: 内蔵 Human68k 風を組み込む。-X による実行ファイル名が必要。
	bool human_mode {};

	// -M: モニタ名
	std::string monitor_opt {};

	// -s,--scale: 画面スケール
	double screen_scale {};

	// -X: ホスト実行ファイル名とその引数
	const char *exec_file {};
	std::string exec_arg {};
	bool load_and_exec {};

 private:
	// ヘルプメッセージを表示する
	void ShowHelp(bool all) const;

	// バージョンを表示する
	void ShowVersion() const;

	// 引数を処理する
	bool ParseOpt(int ac, char *av[]);

	// opt を logopt に追加する (-L オプション用)
	void AddLogopt(const char *opt);

	// -L オプションを処理する
	bool ParseLogopt();

	// ログレベル1つを設定する。
	static bool SetLogopt1(const std::string&, std::string *errmsg);

	// コンパイルされている host netdriver の一覧を表示
	void ShowHostnet() const;

	// CLI/GUI 区別
	bool IsCLI() const { return  is_cli; }
	bool IsGUI() const { return !is_cli; }
	bool is_cli {};

	// ログ
	std::unique_ptr<Logger> logger /*{}*/;

	// 全オブジェクトへのポインタのリスト (所有はしていない)
	std::vector<Object*> objects {};

	// VM
	std::unique_ptr<VM> pVM /*{}*/;

	// 設定項目
	std::unique_ptr<Config> pConfig /*{}*/;

	// VM 種別
	VMType vmtype {};

	// モニタマネージャ
	std::unique_ptr<MonitorManager> pMonitorManager /*{}*/;

	// -c: VM ディレクトリ or 設定ファイル
	std::string vmdir {};
	std::string vmfile {};

	// -C: ログをコンソールに出力
	bool log_to_console {};

	// -L: ログ指定文字列
	std::string logopt {};

	// --show-config: 設定内容を表示 (1なら通常、2ならall)
	int show_config {};

	// -V: パラメータ指定
	std::vector<std::string> config_options {};
};

// グローバルインスタンス
extern MainApp gMainApp;
