//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// M88200 CMMU の I/O デバイス部分
//

// CMMU が確定しない場合はこのデバイスのログを使う。
// CMMU が確定したらそっちのデバイスログから出力する。

#include "cmmu.h"
#include "mpu88xx0.h"

// グローバル参照用
CMMUDevice *gCMMU;

// コンストラクタ
CMMUDevice::CMMUDevice()
	: inherited("CMMUIOSpace")
{
}

// デストラクタ
CMMUDevice::~CMMUDevice()
{
	gCMMU = NULL;
}

bool
CMMUDevice::Init()
{
	// CMMU をポイントしておく
	cmmus[7] = &gMPU88xx0->GetCPU()->cmmu[0];
	cmmus[6] = &gMPU88xx0->GetCPU()->cmmu[1];

	return true;
}

// アドレスデコーダ
/*static*/ inline uint32
CMMUDevice::Decoder(uint32 addr)
{
	uint32 offset = addr & 0xfff;

	// BWP* の A5 はデコードされてない
	if (0x420 <= offset && offset < 0x440) {
		offset &= ~0x20;
	}
	// CDP*,CTP* の A4,A5 はデコードされていない
	if (0x800 <= offset && offset < 0x880) {
		offset &= ~0x30;
	}

	return offset;
}

uint64
CMMUDevice::Read32(uint32 addr)
{
	uint32 id = (addr >> 12) & 0xff;

	m88200 *cmmu = GetCMMU(id);
	if (cmmu == NULL) {
		putlog(3, "Read not-equipped CMMU[%d] $%08x", id, addr);
		return 0xffffffff;
	}

	uint32 offset = Decoder(addr);
	uint32 data;
	switch (offset) {
	 case IDR:
		data = cmmu->GetIDR();
		cmmu->putlogf(2, lstr("IDR  -> $%08x", data));
		return data;

	 case SCR:
		data = cmmu->GetSCR();
		cmmu->putlogf(2, lstr("SCR  -> $%08x", data));
		return data;

	 case SSR:
		data = cmmu->GetSSR();
		cmmu->putlogf(2, lstr("SSR  -> $%08x", data));
		return data;

	 case SAR:
		data = cmmu->GetSAR();
		cmmu->putlogf(2, lstr("SAR  -> $%08x", data));
		return data;

	 case SCTR:
		data = cmmu->GetSCTR();
		cmmu->putlogf(2, lstr("SCTR -> $%08x", data));
		return data;

	 case PFSR:
		data = cmmu->GetPFSR();
		cmmu->putlogf(2, lstr("PFSR -> $%08x", data));
		return data;

	 case PFAR:
		data = cmmu->GetPFAR();
		cmmu->putlogf(2, lstr("PFAR -> $%08x", data));
		return data;

	 case SAPR:
		data = cmmu->GetSAPR();
		cmmu->putlogf(2, lstr("SAPR -> $%08x", data));
		return data;

	 case UAPR:
		data = cmmu->GetUAPR();
		cmmu->putlogf(2, lstr("UAPR -> $%08x", data));
		return data;

	 case BWP0 ... BWP7:
		// BWP は書き込み専用 (何が読めるだろうか)
		return 0xffffffff;

	 case CDP0 ... CDP3:
	 case CTP0 ... CTP3:
		break;

	 case CSSP:
		data = cmmu->GetCSSP();
		cmmu->putlogf(2, lstr("CSSP -> $%08x", data));
		return data;

	 default:
		break;
	}
	putlog(0, "Read $%08x (NOT IMPLEMENTED)", addr);
	return 0xffffffff;
}

uint64
CMMUDevice::Write32(uint32 addr, uint32 data)
{
	uint32 id = (addr >> 12) & 0xff;

	m88200 *cmmu = GetCMMU(id);
	if (cmmu == NULL) {
		putlog(3, "Write not-equipped CMMU[%d] $%08x <- $%08x", id, addr, data);
		return 0;
	}

	uint32 offset = Decoder(addr);
	uint n;
	switch (offset) {
	 case IDR:
	 {
		uint32 newid = data >> 24;
		if (id != newid) {
			cmmu->putlogf(1, lstr("IDR  <- $%08x (set ID $%02x)", data, newid));
			cmmu->SetID(newid);
		} else {
			cmmu->putlogf(2, lstr("IDR  <- $%08x", data));
		}
		return 0;
	 }

	 case SCR:
		cmmu->SetSCR(data);
		return 0;

	 case SSR:
		cmmu->putlogf(1, lstr("SSR  <- $%08x", data));
		cmmu->SetSSR(data);
		return 0;

	 case SAR:
		cmmu->putlogf(2, lstr("SAR  <- $%08x", data));
		cmmu->SetSAR(data);
		return 0;

	 case SCTR:
		cmmu->SetSCTR(data);
		return 0;

	 case PFSR:
		cmmu->SetPFSR(data);
		return 0;

	 case PFAR:
		cmmu->SetPFAR(data);
		return 0;

	 case SAPR:
		cmmu->SetSAPR(data);
		return 0;

	 case UAPR:
		cmmu->SetUAPR(data);
		return 0;

	 case BWP0 ... BWP7:
		n = (offset - BWP0) / 4;
		assert(n < 8);
		cmmu->SetBWP(n, data);
		return 0;

	 case CDP0 ... CDP3:
	 case CTP0 ... CTP3:
		break;

	 case CSSP:
		cmmu->SetCSSP(data);
		return 0;

	 default:
		break;
	}
	putlog(0, "Write $%08x <- $%08x (NOT IMPLEMENTED)", addr, data);
	VMPANIC("not impl");
	return 0;
}

// Peek8() の下請け。32ビットで処理して後でバイトに分割するほうが楽。
uint32
CMMUDevice::Peek32(uint32 addr) const
{
	uint32 id = (addr >> 12) & 0xff;

	m88200 *cmmu = GetCMMU(id);
	if (cmmu == NULL) {
		return 0xffffffff;
	}

	uint32 offset = Decoder(addr);
	switch (offset) {
	 case IDR:
		return cmmu->GetIDR();

	 case SCR:
		return cmmu->GetSCR();

	 case SSR:
		return cmmu->GetSSR();

	 case SAR:
		return cmmu->GetSAR();

	 case SCTR:
		return cmmu->GetSCTR();

	 case PFSR:
		return cmmu->GetPFSR();

	 case PFAR:
		return cmmu->GetPFAR();

	 case SAPR:
		return cmmu->GetSAPR();

	 case UAPR:
		return cmmu->GetUAPR();

	 case BWP0 ... BWP7:
		return 0xffffffff;

	 case CDP0 ... CDP3:
	 case CTP0 ... CTP3:
		break;

	 case CSSP:
		return cmmu->GetCSSP();

	 default:
		break;
	}
	return 0xffffffff;
}

uint64
CMMUDevice::Peek8(uint32 addr)
{
	uint32 data = Peek32(addr & ~3U);
	switch (addr & 3) {
	 case 0:
		return data >> 24;
	 case 1:
		return (data >> 16) & 0xff;
	 case 2:
		return (data >>  8) & 0xff;
	 case 3:
		return data & 0xff;
	}
	__unreachable();
}
