//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// LUNA ビットマッププレーン
//

// SX-9100 のほうはこう書いてあってこっちが正しいようだ。
//      +00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f
// b100 |RF|        |BS|        |共 通 B M P|プレーン#0 |
// b110 |プレーン#1 |プレーン#2 |プレーン#3 |プレーン#4 |
// b120 |プレーン#5 |プレーン#6 |プレーン#7 |FC|        |
// b130 |F0|        |F1|        |F2|        |F3|        |
// b140 |F4|        |F5|        |F6|        |F7|        |
//
// SX-7100 のほうはこう書いてあるがこうはなっていないようだ。
//      +00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f
// b100 |RF|        |BS|        |共 通 B M P|FC|        |
// b110 |プレーン#0 |プレーン#1 |プレーン#2 |プレーン#3 |
// b120 |プレーン#4 |プレーン#5 |プレーン#6 |プレーン#7 |
// b130 |F0|        |F1|        |F2|        |F3|        |
// b140 |F4|        |F5|        |F6|        |F7|        |
//
// RF: RFCNT (4バイト)
// BS: BMSEL (4バイト)
// FC: 共通ファンクションセット
// Fn: プレーン#n ファンクションセット
// 1区画が 64KB。空白のところは未調査。
// 未装着のプレーン(#4..#7とか)は 0xff が読めるようだ。
// ファンクションセットが 64バイトで折り返してるか未調査だが、
// b130'0000 も b133'0000 も 0xff が読めるのでこの範囲全域でミラーかも。
// b150'0000 以降(b1ff'ffffまで) は BUSERR のようだ。

// Lunafb の VRAM はロングワード単位でホストバイトオーダ配置なので、
// バイトアクセス、ワードアクセス時は HLB(), HLW() マクロを使用のこと。

#include "lunafb.h"
#include "bt454.h"
#include "builtinrom.h"
#include "config.h"
#include "mpu.h"

// コンストラクタ
LunafbDevice::LunafbDevice()
	: inherited("Lunafb", 2048, 1024)
{
	devaddr = 0xb1000000;

	modify_mask = 0xffff;

	monitor.func = ToMonitorCallback(&LunafbDevice::MonitorUpdate);
	monitor.SetSize(42, 11);
	monitor.Regist(ID_MONITOR_LUNAVC);
}

// デストラクタ
LunafbDevice::~LunafbDevice()
{
}

// 初期化。
//
// 呼び出し順が際どいので注意。Lunafb::Init() と BT454::Init() とが相互に
// 参照しているが、このうち Lunafb::Init() が呼んでる BT454::GetHostPalette()
// は BT454::Init() 実行前でも取得できる。逆に BT454::Init() が呼んでる
// Lunafb::GetPlaneCount() は Lunafb::Init() の後でないと行えないので、
// vm_luna.cpp での初期化順は PlaneVRAM -> BT454 の順を維持すること。
bool
LunafbDevice::Init()
{
	const ConfigItem& item = gConfig->Find("luna-video-plane");
	if (item.TryInt(&nplane) == false) {
		item.Err();
		return false;
	}
	if (nplane != 1 && nplane != 4) {
		item.Err("must be 1 or 4");
		return false;
	}

	plane_end = plane0base + 0x40000 * nplane;
	fcset_end = fcset0base + 0x40000 * nplane;

	// プレーン数によってメモリを確保
	mem.reset(new uint8 [0x40000 * nplane]);

	// ホストパレットを取得。
	palette = gBT454->GetHostPalette();

	return true;
}

// 電源オン/リセット
void
LunafbDevice::ResetHard(bool poweron)
{
	if (poweron) {
		// XXX 実際は不定値だと思うけど、観測手段がないのでとりあえず。
		memset(&mem[0], 0xff, 0x40000 * nplane);
	}

	// 厳密にどの時点でリセットされるのかは知らないけど
	bmsel = 0;

	// XXX 初期値の観測は困難なので適当
	rfcnt = 0;
	xscroll = 0;
	yscroll = 0;

	for (int i = 0; i < MAX_PLANES; i++) {
		fcset[i] = 5;
		fcmask[i] = 0xffffffff;
	}

	// XXX ここ?
	Invalidate();
}

uint64
LunafbDevice::Read8(uint32 addr)
{
	// XXX とりあえず RAM と同じウェイトを入れておく (未調査)
	gMPU->AddCycle(1);

	if (plane0base <= addr && addr < plane_end) {
		// プレーンの読み出し
		return mem[HLB(addr - plane0base)];
	}
	// 残りは書き込み専用または不在
	putlog(1, "Read not-readable $%08x.B", addr);
	return 0xff;
}

uint64
LunafbDevice::Read16(uint32 addr)
{
	// XXX とりあえず RAM と同じウェイトを入れておく (未調査)
	gMPU->AddCycle(1);

	if (plane0base <= addr && addr < plane_end) {
		// プレーンの読み出し
		return *(uint16 *)&mem[HLW(addr - plane0base)];
	}
	// 残りは書き込み専用または不在
	putlog(1, "Read not-readable $%08x.W", addr);
	return 0xffff;
}

uint64
LunafbDevice::Read32(uint32 addr)
{
	// XXX とりあえず RAM と同じウェイトを入れておく (未調査)
	gMPU->AddCycle(1);

	if (plane0base <= addr && addr < plane_end) {
		// プレーンの読み出し
		return *(uint32 *)&mem[addr - plane0base];
	}
	// 残りは書き込み専用または不在
	putlog(1, "Read not-readable $%08x.L", addr);
	return 0xffffffff;
}

uint64
LunafbDevice::Write8(uint32 addr, uint32 data)
{
	// XXX とりあえず RAM と同じウェイトを入れておく (未調査)
	gMPU->AddCycle(1);

	switch (addr) {
	 case 0xb1000000:
	 case 0xb1000001:
	 case 0xb1000002:
	 case 0xb1000003:
		// このロングワード内へのバイトアクセスはロングワードにミラーされる
		data = (data << 24) | (data << 16) | (data << 8) | data;
		return Write32(addr & ~3, data);
	 case 0xb1040000:
	 case 0xb1040001:
	 case 0xb1040002:
	 case 0xb1040003:
		// BMSEL はバイトアクセスできる?
		putlog(0, "BMSEL $%08x <- $%02x (NOT IMPLEMENTED)", addr, data);
		return 0;
	}

	if (addr < 0xb1080000) {
		// 空き。どうなる?
		putlog(0, "Vacant $%08x <- $%02x (NOT IMPLEMENTED)", addr, data);
		return 0;
	}
	if (addr < plane0base) {
		// 共通ビットマッププレーン
		WriteCommonBitmap8(addr, data);
		return 0;
	}

	if (__predict_true(addr < plane_end)) {
		// 存在するビットマッププレーン
		uint32 reladdr = addr - plane0base;
		uint plane = reladdr / 0x40000;
		uint32 offset = reladdr % 0x40000;

		putlog(2, "Plane%d $%08x <- $%02x", plane, addr, data);
		WritePlane8(plane, offset, data);
		SetDirty(offset);
		return 0;
	}

	if (addr < plane7end) {
		// HW 設定によっては存在しないプレーン
		uint32 reladdr = addr - plane0base;
		uint plane = reladdr / 0x40000;
		putlog(3, "Not-equipped Plane%d $%08x <- $%02x", plane, addr, data);
		return 0;
	}

	// XXX ファンクションセットはバイトアクセスだとどうなる?
	putlog(0, "$%08x <- $%02x (NOT IMPLEMENTED)", addr, data);
	return 0;
}

uint64
LunafbDevice::Write16(uint32 addr, uint32 data)
{
	// XXX とりあえず RAM と同じウェイトを入れておく (未調査)
	gMPU->AddCycle(1);

	switch (addr) {
	 case 0xb1000000:
	 case 0xb1000002:
		// このロングワード内へのワードアクセスはロングワードにミラーされる
		data = (data << 16) | data;
		return Write32(addr & ~3, data);
	 case 0xb1040000:
	 case 0xb1040002:
		// BMSEL はワードアクセスできる?
		putlog(0, "BMSEL $%08x <- $%04x (NOT IMPLEMENTED)", addr, data);
		return 0;
	}

	if (addr < 0xb1080000) {
		// 空き。どうなる?
		putlog(0, "Vacant $%08x <- $%04x (NOT IMPLEMENTED)", addr, data);
		return 0;
	}
	if (addr < plane0base) {
		// 共通ビットマッププレーン
		WriteCommonBitmap16(addr, data);
		return 0;
	}

	if (__predict_true(addr < plane_end)) {
		// 存在するビットマッププレーン
		uint32 reladdr = addr - plane0base;
		uint plane = reladdr / 0x40000;
		uint32 offset = reladdr % 0x40000;

		putlog(2, "Plane%d $%08x <- $%04x", plane, addr, data);
		WritePlane16(plane, offset, data);
		SetDirty(offset);
		return 0;
	}

	if (addr < plane7end) {
		// HW 設定によっては存在しないプレーン
		uint32 reladdr = addr - plane0base;
		uint plane = reladdr / 0x40000;
		putlog(3, "Not-equipped Plane%d $%08x <- $%04x", plane, addr, data);
		return 0;
	}

	// XXX ファンクションセットはワードアクセスだとどうなる?
	putlog(0, "$%08x <- $%04x (NOT IMPLEMENTED)", addr, data);
	return 0;
}

uint64
LunafbDevice::Write32(uint32 addr, uint32 data)
{
	// XXX とりあえず RAM と同じウェイトを入れておく (未調査)
	gMPU->AddCycle(1);

	switch (addr) {
	 case 0xb1000000:
		// RFCNT
		WriteRFCNT(data);
		return 0;

	 case 0xb1040000:
		// BMSEL
		bmsel = data & 0xf;
		putlog(1, "BMSEL <- $%08x", data);
		return 0;
	}

	if (addr < 0xb1080000) {
		putlog(0, "Vacant $%08x <- $%08x (NOT IMPLEMENTED)", addr, data);
		return 0;
	}
	if (addr < plane0base) {
		// 共通ビットマッププレーン
		WriteCommonBitmap32(addr, data);
		return 0;
	}

	if (__predict_true(addr < plane_end)) {
		// 存在するビットマッププレーン
		uint32 reladdr = addr - plane0base;
		uint plane = reladdr / 0x40000;
		uint32 offset = reladdr % 0x40000;

		putlog(3, "Plane%d $%08x <- $%08x", plane, addr, data);
		WritePlane32(plane, offset, data);
		SetDirty(offset);
		return 0;
	}

	if (__predict_false(addr < plane7end)) {
		// HW 設定によっては存在しないプレーン
		uint32 reladdr = addr - plane0base;
		uint plane = reladdr / 0x40000;
		putlog(3, "Not-equipped Plane%d $%08x <- $%08x", plane, addr, data);
		return 0;
	}

	if (addr < 0xb1300000) {
		// 共通ファンクションセット
		uint rop = (addr & 0x3f) / 4;
		putlog(2, "Common FCSet $%08x <- $%08x (%s)",
			addr, data, RopStr(rop));
		WriteCommonFCSet(rop, data);
		return 0;
	}

	if (__predict_true(addr < fcset_end)) {
		// 存在するファンクションセット
		int plane = (addr - 0xb1300000) / 0x40000;
		uint rop = (addr & 0x3f) / 4;
		putlog(2, "FCSet%d $%08x <- $%08x (%s)",
				plane, addr, data, RopStr(rop));
		fcset[plane] = rop;
		fcmask[plane] = data;
		return 0;
	}

	if (addr < fcset7end) {
		// HW 設定によっては存在しないファンクションセット
		int plane = (addr - 0xb1300000) / 0x40000;
		putlog(3, "Not-equipped FCSet%d $%08x <- $%08x", plane, addr, data);
		return 0;
	}

	// 残りは全部空きのはずだが挙動未調査。
	putlog(0, "$%08x <- $%08x (NOT IMPLEMENTED)", addr, data);
	return 0;
}

uint64
LunafbDevice::Peek8(uint32 addr)
{
	if (plane0base <= addr && addr < plane_end) {
		// プレーンの読み出し
		return mem[HLB(addr - plane0base)];
	}
	// 残りは書き込み専用
	return 0xff;
}

void
LunafbDevice::MonitorUpdate(Monitor *, TextScreen& screen)
{
	int i;

	screen.Clear();

	// 0         1         2         3         4
	// 012345678901234567890123456789012345678901
	// RFCNT: $12345678 (X=-123dot, Y=1023dot)
	//
	//    BMSEL=$4 ROP  Mask           RGB    RGB
	// Plane#0 Sel THRU $ffffffff   $0:RGB $8:RGB
	screen.Print(0, 0, "RFCNT: $%08x (X=%ddot, Y=%ddot)",
		rfcnt, xscroll, yscroll);
	screen.Print(3, 2, "BMSEL=$%x ROP  Mask", bmsel);

	for (i = 0; i < nplane; i++) {
		screen.Print(0, 3 + i, "Plane#%d %s %s $%08x",
			i,
			(bmsel & (1 << i)) ? "Sel" : "---",
			RopStr(fcset[i]), fcmask[i]);
	}
	for (; i < MAX_PLANES; i++) {
		screen.Print(0, 3 + i, TA::Disable, "Plane#%d (Not installed)", i);
	}

	// ついでなのでここで BT454 のパレットも表示。
	// 012345
	// $0:RGB
	const uint8 *pal = gBT454->reg.color;
	screen.Puts(29, 1, "<BT454>");
	screen.Puts(32, 2, "RGB    RGB");
	for (i = 0; i < 16; i++) {
		uint rgb =
			((pal[i * 3 + 0] & 0xf0) << 4) +
			((pal[i * 3 + 1] & 0xf0)     ) +
			((pal[i * 3 + 2] & 0xf0) >> 4);
		screen.Print(29 + (i / 8) * 7, (i % 8) + 3, "$%x:%03x", i, rgb);
	}
}

// offset の位置に対応する更新フラグを立てる。
// offset はプレーン先頭から (VRAM 先頭からでも同じ) のバイトオフセット。
inline void
LunafbDevice::SetDirty(uint32 offset)
{
	// VRAM は横 2048 ドット(256バイト) x 縦 1024 ドットで、これを
	// 横 128 ドット(16バイト) x 縦 16 ドットずつのブロックに分割する。
	// ブロック数は横 16個 x 縦 64個。
	//
	//                1                               0
	//        4 3 2 1 0 f e d c b a 9 8 7 6 5 4 3 2 1 0
	//       +-----------+-------+-------+-------+-----+
	// VRAM  |  Y座標 (1024dot)  |  X (256byte)  : 8bit|
	//       +-----------+-------+-------+-------+-----+
	//       :           :       :       :             :
	//       +-----------+-------+-------+-------------+
	// Block | Y(64blks) | 16dot | X(16) |   128dot    | *: X は 16 blocks
	//       +-----------+-------+-------+-------------+
	//        ＼          ＼     |       |
	//          ＼          ＼   |       |
	//            ＼          ＼ |       |
	//               +-----------+-------+
	// Dirty         | Y(64blks) | X(16) |
	//               +-----------+-------+
	//                9 8 7 6 5 4 3 2 1 0

	// x 座標(バイト単位 8 bit)の上位 4 bit
	int x = (offset >> 4) & 0x0f;
	// y 座標(10bit) の上位 6 bit
	int y = (offset >> 8) & 0x3f0;

	// からインデックスを求める
	int b = x + y;

	dirty[b] = 1;

#if defined(PERF_HIT)
	// どのロングワードが更新されたか
	dirty_word[offset >> 2] = true;
#endif
}

// RFCNT への書き込み
void
LunafbDevice::WriteRFCNT(uint32 data)
{
	// レジスタ値
	rfcnt = data;

	// 水平オフセットを計算
	// XXX マニュアルでは -8 まで指定できるように読めるが
	//     指定する方法が分からない。
	// XXX 193 以上を指定するとどうなるのか (とりあえずほっとく)
	//     実機の挙動を目分量で見ると、右にはみ出た部分は
	//     256 ラスタ先のラスタデータを描画しているようだ。
	xscroll = ((rfcnt >> 16) & 0xff) + 8;
	xscroll *= 4;

	// ラスタ(垂直)カウンタ
	yscroll = (rfcnt + 26) & 0x3ff;

	// 全画面更新
	Invalidate();

	putlog(1, "RFCNT <- $%08x (H=%ddot, V=%ddot)", data, xscroll, yscroll);
}

// 共通ビットマッププレーン バイト書き込み
void
LunafbDevice::WriteCommonBitmap8(uint32 addr, uint32 data)
{
	uint32 offset = addr - 0xb1080000;
	putlog(2, "Common Plane $%08x <- $%02x", addr, data);
	for (int i = 0; i < nplane; i++) {
		if ((bmsel & (1 << i))) {
			WritePlane8(i, offset, data);
		}
	}
	SetDirty(offset);
}

// 共通ビットマッププレーン ワード書き込み
void
LunafbDevice::WriteCommonBitmap16(uint32 addr, uint32 data)
{
	uint32 offset = addr - 0xb1080000;
	putlog(2, "Common Plane $%08x <- $%04x", addr, data);
	for (int i = 0; i < nplane; i++) {
		if ((bmsel & (1 << i))) {
			WritePlane16(i, offset, data);
		}
	}
	SetDirty(offset);
}

// 共通ビットマッププレーン ロング書き込み
void
LunafbDevice::WriteCommonBitmap32(uint32 addr, uint32 data)
{
	uint32 offset = addr - 0xb1080000;
	putlog(2, "Common Plane $%08x <- $%08x", addr, data);
	for (int i = 0; i < nplane; i++) {
		if ((bmsel & (1 << i))) {
			WritePlane32(i, offset, data);
		}
	}
	SetDirty(offset);
}

// 共通ファンクションセット設定
void
LunafbDevice::WriteCommonFCSet(uint rop, uint32 data)
{
	for (int i = 0; i < nplane; i++) {
		if ((bmsel & (1 << i))) {
			fcset[i] = rop;
			fcmask[i] = data;
		}
	}
}

// 現在の VRAM の値 vram と、書き込まれた値 data から
// 演算後 VRAM に書き戻すことになる値を返す。
// ここではマスクはまだ評価しない。
/*static*/ uint32
LunafbDevice::RasterOp(uint rop, uint32 vram, uint32 data)
{
	switch (rop) {
	 case ROP::ZERO:
		vram = 0;
		break;
	 case ROP::AND1:
		vram = vram & data;
		break;
	 case ROP::AND2:
		vram = vram & ~data;
		break;
	 case ROP::cmd:
		// 使用不可。本当はここにシリアルモード変更コマンドがあるらしい。

		// ログを出すためにはインスタンスが必要だが、普段は static 関数で
		// 十分なので、ログを出す時だけグローバルインスタンスを参照する。
		if (gPlaneVRAM->loglevel >= 0) {
			gPlaneVRAM->putlogn("ROP 3 is used");
		}
		break;
	 case ROP::AND3:
		vram = ~vram & data;
		break;
	 case ROP::THRU:
		vram = data;
		break;
	 case ROP::EOR:
		vram = vram ^ data;
		break;
	 case ROP::OR1:
		vram = vram | data;
		break;
	 case ROP::NOR:
		vram = ~vram & ~data;
		break;
	 case ROP::ENOR:
		vram = (vram & data) | (~vram & ~data);
		break;
	 case ROP::INV1:
		vram = ~data;
		break;
	 case ROP::OR2:
		vram = vram | ~data;
		break;
	 case ROP::INV2:
		vram = ~vram;
		break;
	 case ROP::OR3:
		vram = ~vram | data;
		break;
	 case ROP::NAND:
		vram = ~vram | ~data;
		break;
	 case ROP::ONE:
		vram = 0xffffffff;
		break;
	 default:
		__unreachable();
	}
	return vram;
}

// ROP 文字列を返す
const char *
LunafbDevice::RopStr(uint rop)
{
	static const char * ropstr[] = {
		"ZERO",
		"AND1",
		"AND2",
		"(3)",
		"AND3",
		"THRU",	// THROUGH
		"EOR",
		"OR1",
		"NOR",
		"ENOR",
		"INV1",
		"OR2",
		"INV2",
		"OR3",
		"NAND",
		"ONE",
	};
	return ropstr[rop];
}

// 個別プレーンへのバイト書き込み
// (dirty は変更しない)
void
LunafbDevice::WritePlane8(uint plane, uint32 offset, uint32 data)
{
	uint32 memoffset = plane * 0x40000 + offset;
	uint32 vram;
	uint32 mask;

	vram = mem[HLB(memoffset)];
	data = RasterOp(fcset[plane], vram, data);
	mask = fcmask[plane] >> (24 - (memoffset & 3) * 8);
	mem[HLB(memoffset)] = (data & mask) | (vram & ~mask);
}

// 個別プレーンへのワード書き込み
// (dirty は変更しない)
void
LunafbDevice::WritePlane16(uint plane, uint32 offset, uint32 data)
{
	uint32 memoffset = plane * 0x40000 + offset;
	uint32 vram;
	uint32 mask;

	vram = *(uint16 *)&mem[HLW(memoffset)];
	data = RasterOp(fcset[plane], vram, data);
	if ((memoffset & 2) == 0) {
		// 上位ワード
		mask = fcmask[plane] >> 16;
	} else {
		// 下位ワード
		mask = fcmask[plane];
	}
	*(uint16 *)&mem[HLW(memoffset)] = (data & mask) | (vram & ~mask);
}

// 個別プレーンへのロングワード書き込み
// (dirty は変更しない)
void
LunafbDevice::WritePlane32(uint plane, uint32 offset, uint32 data)
{
	uint32 memoffset = plane * 0x40000 + offset;
	uint32 vram;
	uint32 mask;

	vram = *(uint32 *)&mem[memoffset];
	data = RasterOp(fcset[plane], vram, data);
	mask = fcmask[plane];
	*(uint32 *)&mem[memoffset] = (data & mask) | (vram & ~mask);
}

// (x, y) 位置のピクセル情報を screen に描画。(ビットマップモニタの情報欄)
void
LunafbDevice::UpdateInfo(TextScreen& screen, int x, int y) const
{
	screen.Clear();

	//0         1         2         3         4         5         6
	//012345678901234567890123456789012345678901234567890123456789012345
	//X=1024 Y=2024: (P0) $b10c0000: $00 (%00000000) ColorCode=$d

	screen.Puts(0, 0, "X=     Y=");
	for (int i = 0; i < nplane; i++) {
		screen.Print(15, i, "(P%d)", i);
	}

	// カーソルが範囲外ならここまで。
	if (x < 0) {
		return;
	}

	screen.Print(2, 0, "%4d Y=%4d", x, y);

	// 1プレーン内のバイトオフセット
	uint32 planeoffset = ((uint)y * 256) + ((uint)x / 8);

	uint cc = 0;
	for (int i = 0; i < nplane; i++) {
		// プレーン0先頭からのバイトオフセット
		uint32 totaloffset = 0x40000 * i + planeoffset;

		uint8 data = mem[HLB(totaloffset)];
		screen.Print(15, i, "(P%d) $%08x: $%02x (%%",
			i, plane0base + totaloffset, data);

		// ビットパターン
		for (int m = 0x80; m != 0; m >>= 1) {
			screen.Putc((data & m) ? '1' : '0');
		}
		screen.Putc(')');

		// カーソル位置をハイライト
		screen.Locate(37 + (x % 8), i);
		screen.SetAttr(TA::Em);

		// カラーコードを計算
		if ((data & (0x80U >> (x % 8))) != 0) {
			cc |= (1U << i);
		}
	}

	screen.Print(47, 0, "ColorCode=$%x", cc);
}


// エミュレータによる画面出力
//
// VM 内からの操作ではなく、エミュレータ本体から画面を操作する
// エミュレーション ROM によるモニタコンソール用の機能。
// サービスを使いたい側 (EmuROM など) は EmuInitScreen() をコールすると
// (現在のパレットのまま) 画面をクリアする。
// 以降は EmuPutc() で文字を表示する。

// エミュレータ出力用に画面を初期化する
void
LunafbDevice::EmuInitScreen()
{
	// 同時書き込み設定
	bmsel = (1 << nplane) - 1;

	// ROP を初期化
	WriteCommonFCSet(ROP::THRU, 0xffffffff);

	// 全画面をクリア
	EmuFill(0, 0, 2048, 1024, 0);
}

// 矩形領域を data で埋める。
void
LunafbDevice::EmuFill(uint x, uint y, uint w, uint h, uint32 data)
{
	// ロングワードアラインしていないケースは対応しない
	assert(x % 32 == 0);
	assert(w % 32 == 0);

	for (uint py = y; py < y + h; py++) {
		uint32 addr = 0xb1080000 + (x / 32) * 4 + py * 256;
		for (uint px = x; px < x + w; px += 32) {
			WriteCommonBitmap32(addr, data);
			addr += 4;
		}
	}
}

// ピクセル座標 x, y を左上として文字 c を出力。
// 文字が VRAM からはみ出ないようにすること。
void
LunafbDevice::EmuPutc(uint x, uint y, uint c)
{
	EmuPutc(x, y, c, ROP::THRU);
}

// ピクセル座標 x, y を左上として文字 c を出力。
// 文字が VRAM からはみ出ないようにすること。
void
LunafbDevice::EmuPutc(uint x, uint y, uint c, int rop)
{
	uint32 addr = 0xb1080000 + (x / 32) * 4 + y * 256;
	uint32 bofs = x % 32;

	// マスク設定。上から 12bit にマスクを置いといて..
	union64 mask {};
	mask.h = 0xfff00000;

	// うまい具合にシフト
	mask.q >>= bofs;

	// その文字のフォントデータ上でのアドレス
	const uint8 * const fontaddr = Builtin::CGROM12x24(c & 0xff);

	for (int i = 0; i < 24; i++) {
		// その文字のフォントデータ。
		// データは MSB 側 12bit に入っているのでこれを 64bit MSB 詰めにする
		union64 data {};
		data.h = be16toh(*(const uint16*)(fontaddr + i * 2)) << 16;

		// うまい具合にシフト
		data.q >>= bofs;

		WriteCommonFCSet(rop, mask.h);
		WriteCommonBitmap32(addr, data.h);
		if (mask.l) {
			WriteCommonFCSet(rop, mask.l);
			WriteCommonBitmap32(addr + 4, data.l);
		}
		addr += 256;	// 1ライン進める
	}

	// ROP を元に戻す
	WriteCommonFCSet(ROP::THRU, 0xffffffff);
}

// 画面を縦スクロールさせる。上スクロールのみサポート。
void
LunafbDevice::EmuScrollY(uint dy, uint sy, uint h)
{
	assert(dy < sy);

	uint32 da = dy * 256;
	uint32 sa = sy * 256;

	for (uint y = 0; y < h; y++) {
		for (int i = 0; i < nplane; i++) {
			memcpy(&mem[da + i * 0x40000], &mem[sa + i * 0x40000], 160);
		}
		da += 256;
		sa += 256;
	}

	Invalidate();
}
