//
// nono
// Copyright (C) 2022 nono project
// Licensed under nono-license.txt
//

//
// プレーン VRAM (垂直 VRAM)
// Lunafb と X68k TVRAM の共通部分
//

#pragma once

#include "device.h"
#include "renderer.h"
#include <vector>

class PlaneVRAMDevice : public IODevice
{
	using inherited = IODevice;

 protected:
	// ショートカット
	static const int BLKX = Renderer::BLKX;
	static const int BLKY = Renderer::BLKY;

 public:
	PlaneVRAMDevice(const std::string& objname_, int width_, int height_);
	virtual ~PlaneVRAMDevice() override;

	void ResetHard(bool poweron) override;

	// レンダリング時に使うテーブルを事前に計算する。
	void InitDeptable();

	// プレーン数を取得
	int GetPlaneCount() const { return nplane; }

	// 更新情報を取得
	ModifyInfo GetModify() const;

	// 更新フラグをクリア
	void ClearDirty();

	// 画面を合成する
	bool Render(BitmapRGB& dst, const ModifyInfo& modify);

	// 合成ビットマップを取得する
	const BitmapI8& GetComposite() const { return composite; }

	// 座標位置の情報を出力 (GUI 用)
	virtual void UpdateInfo(TextScreen&, int x, int y) const = 0;

#if defined(PERF_HIT)
	// ブロックのヒット効率(?)を調べる
	std::array<bool, 65536> dirty_word {};
#endif

 protected:
	// 全画面の更新フラグを立てる
	void Invalidate();

	// 垂直 VRAM から BitmapI8 バッファを作成する
	void RenderVRAMToComposite(const ModifyInfo& modify);

	// BitmapI8 バッファから BitmapRGB を作成する。
	void RenderCompositeToRGB(BitmapRGB& dst, const ModifyInfo& modify);

	// ビットマッププレーン。
	// メモリを確保する時のサイズ計算のため型を uint8 にしているが、
	// 運用はロングワードでホストバイトオーダー。
	// Lunafb では 1枚 256KBで 1 or 4枚。TVRAM では 512KB。
	std::unique_ptr<uint8[]> mem {};

	// プレーン数 (TVRAM なら 4 固定、LUNA は設定による)
	int nplane {};

	// テキスト画面の合成ビットマップ
	BitmapI8 composite {};

	// 更新フラグ。
	// Lunafb では 1024、TVRAM では 512個。
	std::vector<uint8> dirty {};

	// ModifyInfo.bits に対するマスク。横方向のブロック数によって決まり、
	// Lunafb では 0xffff、TVRAM では 0xff00。
	uint16 modify_mask {};

	// レンダリング時に使う変換テーブル
	std::unique_ptr<uint64[]> deptable {};

	// ホストパレット
	const ColorXBGR *palette {};

	// スクロール
	int xscroll {};
	int yscroll {};
};

extern PlaneVRAMDevice *gPlaneVRAM;
