//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// PROM (LUNA-I, LUNA88K 共通)
//

// IODevice
//    |
//    +-- ROMDevice (LoadROM()、ウェイト、マスク等を持つ)
//    |     |
//    |     |  +------------+
//    |     +--| PROMDevice |  (LUNA* の PROM)
//    |     |  +------------+
//    |     |
//    |     +-- IPLROM1Device (X680x0 の IPLROM 後半)
//    |     +-- IPLROM2Device (X680x0 の IPLROM 前半)
//    |     +-- CGROMDevice   (X680x0 の CGROM)
//    |     +-- LunaPROMEmuDevice   (LUNA PROM エミュレーションの共通部分)
//    |           |
//    |           +-- Luna1PROMEmuDevice   (LUNA1 の PROM エミュレーション)
//    |           +-- Luna88kPROMEmuDevice (LUNA88K の PROM エミュレーション)
//    |
//    |   +-------------+
//    +-- | PROM0Device | (LUNA* のブートページ切り替え用プロキシ)
//    |   +-------------+
//    |
//    +-- IPLROM0Device   (X680x0 のブートページ切り替え用プロキシ)

#include "prom.h"
#include "config.h"
#include "ethernet.h"
#include "mainapp.h"
#include "vm.h"

// グローバル参照用
PROM0Device *gPROM0;
PROMDevice *gPROM;

//
// PROM0
//
// PROM0 は PROM のブートページ担当。
// これ自身はメモリ領域を持たず PROM へのプロキシとして動作する。
//
// リセット後最初の ROM 領域への書き込みアクセスでブートページを RAM に
// 切り替える。
// LUNA のブートページのアドレス範囲は不明なので、とりあえず devtable の
// 1区画である 16MB ($0000'0000..$00ff'ffff) とする。実害はない。

// コンストラクタ
PROM0Device::PROM0Device()
	: inherited("PROM(BOOT)")
{
	ClearAlias();
	AddAlias("PROMBOOT");
}

// デストラクタ
PROM0Device::~PROM0Device()
{
	gPROM0 = NULL;
}

bool
PROM0Device::Init()
{
	prom = gPROM;
	return true;
}

// リセット
void
PROM0Device::ResetHard(bool poweron)
{
	gVM->SwitchBootPageToROM(this);
}

// アドレスデコーダ
/*static*/ inline uint32
PROM0Device::Decoder(uint32 addr)
{
	// ここはブートページか ROM かの比較だけで十分 (それ以外はここに来ない)
	if (addr < 0x0100'0000) {
		addr += 0x4100'0000;
	}
	return addr;
}

uint64
PROM0Device::Read8(uint32 addr)
{
	addr = Decoder(addr);
	return prom->Read8(addr);
}

uint64
PROM0Device::Read16(uint32 addr)
{
	addr = Decoder(addr);
	return prom->Read16(addr);
}

uint64
PROM0Device::Read32(uint32 addr)
{
	addr = Decoder(addr);
	return prom->Read32(addr);
}

// LUNA では PROM へのライトアクセスでブートページを RAM に切り替えている。
// - LUNA-I では最初に $41000000.B への書き込みがあるが、
//   それ以降も $41000000.L への書き込みが(複数回)あるようだ。
// - LUNA88K では $4100000c.L への書き込みがある。
//
// よってアドレスを定めず PROM 全域への書き込みで最初の1回だけ発動する
// ようにしておく。サイズも今の所ワード(16bit)アクセスは確認してないけど
// 統一的に対処しておく。
uint64
PROM0Device::Write()
{
	gVM->SwitchBootPageToRAM(this);

	// バスエラーは起きない
	return 0;
}

uint64
PROM0Device::Write8(uint32 addr, uint32 data)
{
	return Write();
}

uint64
PROM0Device::Write16(uint32 addr, uint32 data)
{
	return Write();
}

uint64
PROM0Device::Write32(uint32 addr, uint32 data)
{
	return Write();
}

uint64
PROM0Device::Peek8(uint32 addr)
{
	addr = Decoder(addr);
	return prom->Peek8(addr);
}


//
// PROM
//

// コンストラクタ
PROMDevice::PROMDevice()
	: inherited("PROM")
{
	devaddr = 0x41000000;

	// LUNA では ROM への書き込みは何も起きない
	write_op = 0;
}

// デストラクタ
PROMDevice::~PROMDevice()
{
	gPROM = NULL;
}

bool
PROMDevice::Init()
{
	VMType vmtype = gMainApp.GetVMType();

	switch (vmtype) {
	 case VMType::LUNA1:
		return InitLuna1();

	 case VMType::LUNA88K:
		return InitLuna88k();

	 default:
		PANIC("unknown vmtype %d", (int)vmtype);
	}
}

// LUNA-I の場合
bool
PROMDevice::InitLuna1()
{
	devlen = 128 * 1024;
	mask = devlen - 1;

	// ファイル名が存在してる時しかここに来ない
	const ConfigItem& item = gConfig->Find("prom-image");
	if (LoadROM(item.AsString().c_str(), devlen) == false) {
		return false;
	}

	// XXX アクセスウェイトは未調査。とりあえず RAM と同程度にしておく
	wait = 1;

	// ROM バージョンを確認。確認されてるのは
	// V4.02 (Wed Oct 12 17:07:11 1988)
	// V4.22 (Thu Jul 27 11:45:42 1989)
	// V4.25
	//
	// とりあえず V4.22 は +$00d390 からにタイムスタンプ文字列があるので
	// それで判別してみる。
	//00d390 54 68 75 20 4a 75 6c 20 32 37 20 31 31 3a 34 35 |Thu Jul 27 11:45|
	//00d3a0 3a 34 32 20 31 39 38 39 00 00 00 00 4e 56 fd f8 |:42 1989....NV..|

	if (strcmp((const char *)&mem[0xd390], "Thu Jul 27 11:45:42 1989") == 0) {
		// V4.22
		putmsg(1, "V4.22 detected");

		// MAC アドレスは ROM に書き込まれているので、ここで上書き
		if (strcmp((const char *)&mem[0x1ffd8], "ENADDR") != 0) {
			warnx("ENADDR not found; unknown ROM V4.22 variants?");
			return false;
		}

		// MAC アドレスは12桁の16進数文字列として書き込まれている。例えば
		// 前半の 00:00:0A 部分は 30 30 30 30 30 41 "00000A" という感じ。
		const macaddr_t macaddr = gEthernet->GetMacAddr();
		memcpy(&imagebuf[0x1ffe0], macaddr.ToString().c_str(), 12);

		// SCSI デバッグビット
		bool scsi_debug_bit = gConfig->Find(".prom-scsi-debug").AsInt();
		imagebuf[0x4100981d - 0x41000000] = (scsi_debug_bit) ? 0x01 : 0;
	} else {
		// 他の ROM は何も分からないので何もしない。
		putmsg(0, "Unknown PROM Version");
	}

	return true;
}

// LUNA88K の場合
bool
PROMDevice::InitLuna88k()
{
	devlen = 256 * 1024;
	mask = devlen - 1;

	// ファイル名が存在してる時しかここに来ない
	const ConfigItem& item = gConfig->Find("prom-image");
	if (LoadROM(item.AsString().c_str(), devlen) == false) {
		return false;
	}

	// XXX アクセスウェイトは未調査。とりあえず RAM と同程度にしておく
	wait = 3;

	return true;
}
