//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// ROM の基本クラス
//

// IODevice
//    |
//    |  +-----------+
//    +--| ROMDevice | (LoadROM()、ウェイト、マスク等を持つ)
//    |  +-----------+
//    |     |
//    |     +-- PROMDevice    (LUNA* の PROM)
//    |     +-- IPLROM1Device (X680x0 の IPLROM 後半)
//    |     +-- IPLROM2Device (X680x0 の IPLROM 前半)
//    |     +-- CGROMDevice   (X680x0 の CGROM)
//    |     +-- LunaPROMEmuDevice (LUNA PROM エミュレーションの共通部分)
//    |           |
//    |           +-- Luna1PROMEmuDevice   (LUNA1 の PROM エミュレーション)
//    |           +-- Luna88kPROMEmuDevice (LUNA88K の PROM エミュレーション)
//    |
//    +-- PROM0Device   (LUNA* のブートページ切り替え用プロキシ)
//    +-- IPLROM0Device (X680x0 のブートページ切り替え用プロキシ)

#include "rom.h"
#include "autofd.h"
#include "mainapp.h"
#include "mpu.h"
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>

//
// ROM 共通部
//

ROMDevice::ROMDevice(const std::string& objname_)
	: inherited(objname_)
{
	mask = 0xffffffff;
}

ROMDevice::~ROMDevice()
{
}

// name の ROM を探してロードする。
// name はパスを含まないファイル名のみ。パスはこちらで探す。
// 失敗した場合はエラーメッセージを表示して false を返す。
bool
ROMDevice::LoadROM(const char *name, int size)
{
	struct stat st;

	assert(size > 0);

	// ファイル名
	filename = gMainApp.SearchFile(name);
	if (filename.empty()) {
		warnx("%s not found", name);
		return false;
	}

	// メモリを確保
	imagebuf.reset(new uint8 [size]);
	if ((bool)imagebuf == false) {
		warnx("ROM \"%s\": Cannot allocate memory (%d bytes)",
			filename.c_str(), size);
		return false;
	}

	// オープン
	autofd fd = open(filename.c_str(), O_RDONLY);
	if (fd < 0) {
		warn("ROM \"%s\"", filename.c_str());
		return false;
	}

	// ファイルサイズを調べる
	if (fstat(fd, &st) < 0) {
		warn("ROM \"%s\" stat failed", filename.c_str());
		return false;
	}

	if (st.st_size < size) {
		warnx("ROM \"%s\" is too short (filesize is expected %d but %jd)",
			filename.c_str(), size, (intmax_t)st.st_size);
		return false;
	}
	if (st.st_size > size) {
		// 大きすぎるほうはワーニングだけでいいか
		warnx("%s is too large (ignore %zd bytes)",
			filename.c_str(), (ssize_t)st.st_size - size);
	}

	// 読み込む
	ssize_t n = read(fd, &imagebuf[0], size);
	if (n < 0) {
		warn("ROM \"%s\" read failed", filename.c_str());
		return false;
	}
	if (n < size) {
		warnx("ROM \"%s\" read too short", filename.c_str());
		return false;
	}

	fd.Close();

	mem = &imagebuf[0];
	return true;
}

uint64
ROMDevice::Read8(uint32 addr)
{
	gMPU->AddCycle(wait);
	uint32 data;
	data = mem[addr & mask];
	putlog(4, "$%08x.B -> $%02x", addr, data);
	return data;
}

uint64
ROMDevice::Read16(uint32 addr)
{
	gMPU->AddCycle(wait);
	uint16 data;
	data = be16toh(*(const uint16 *)&mem[addr & mask]);
	putlog(4, "$%08x.W -> $%04x", addr, data);
	return data;
}

uint64
ROMDevice::Read32(uint32 addr)
{
	gMPU->AddCycle(wait);
	uint32 data;
	data = be32toh(*(const uint32 *)&mem[addr & mask]);
	putlog(4, "$%08x.L -> $%08x", addr, data);
	return data;
}

uint64
ROMDevice::Write8(uint32 addr, uint32 data)
{
	gMPU->AddCycle(wait);	// ?
	putlog(4, "$%08x.B <- $%02x", addr, data);
	return write_op;
}

uint64
ROMDevice::Write16(uint32 addr, uint32 data)
{
	gMPU->AddCycle(wait);	// ?
	putlog(4, "$%08x.W <- $%04x", addr, data);
	return write_op;
}

uint64
ROMDevice::Write32(uint32 addr, uint32 data)
{
	gMPU->AddCycle(wait);	// ?
	putlog(4, "$%08x.L <- $%08x", addr, data);
	return write_op;
}

uint64
ROMDevice::Peek8(uint32 addr)
{
	return mem[addr & mask];
}

// ファイル名を返す。なければ NULL を返す。
const char *
ROMDevice::GetFilename() const
{
	if (filename.empty()) {
		return NULL;
	} else {
		return filename.c_str();
	}
}
