//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// LUNA シリーズの ROM エミュレーション共通部分
//

#pragma once

#include "rom.h"

class LunaPROMEmuDevice : public ROMDevice
{
	using inherited = ROMDevice;

 protected:
	// このデバイス先頭アドレス
	static const uint32 baseaddr		= 0x41000000;

	// 謎の I/O 空間
	static const uint32 ROMIO_BASE		= 0x41010000;
	static const uint32 ROMIO_INIT		= (ROMIO_BASE + 0);
	static const uint32 ROMIO_KEYIN		= (ROMIO_BASE + 4);
	static const uint32 ROMIO_CLOSE		= (ROMIO_BASE + 8);
	static const uint32 ROMIO_ENTRY		= (ROMIO_BASE + 12);

	// 文字コード
	// 概ね ASCII だが制御コード部分は必要な文字だけを適当に割り当てる。
	static const char BS		= 0x08;
	static const char LF		= 0x0a;
	static const char UP		= 0x1c;
	static const char LEFT		= 0x1d;
	static const char RIGHT		= 0x1e;
	static const char DOWN		= 0x1f;

	// ブート情報格納用
	struct bootinfo_t {
		int dkunit {};
		int dkpart {};
		std::string dkfile {};
	};

 public:
	LunaPROMEmuDevice();
	virtual ~LunaPROMEmuDevice() override;

	bool Init() override;

	uint64 Read32(uint32 addr) override;

 protected:
	virtual uint32 ROM_Init() = 0;
	virtual uint32 ROM_Entry() = 0;
	virtual void ROM_Close();
	virtual void GetChar(char);
	virtual void Command() = 0;

	void InitDevice();
	void InitScreen();
	void PrintTitle(const char *machine_name);
	void ClearPrompt();
	void Putc(uint ch);
	void Print(const std::string& s);
	void Print(const char *fmt, ...) __printflike(2, 3);

	// NVRAM
	uint8 CalcNVRAMCsum() const;
	void WriteNVRAMCsum();
	bool VerifyNVRAM() const;

	uint32 LoadHostFile();
	uint32 LoadGuestFile(const bootinfo_t&);

	// エラーメッセージの受け渡し用
	// 基本的にエラーが起きた時だけ更新するので、ここにエラーメッセージを
	// 返す可能性がある関数をコールする前に呼び出し側が clear() しておき、
	// 関数から戻ったら empty() かどうか確認する、errno みたいな使い方。
	std::string errmsg {};

	// 入力バッファ
	std::string inputbuf {};

	// 現在のプロンプト文字列
	std::string prompt {};

	// デフォルトプロンプト。
	const std::string default_prompt = ">";

	// プロンプトのある行(最上行が 0)
	int prompt_y {};

	// カーソル位置(入力バッファ先頭が0)
	int inputpos {};

	int screen_w {};		// テキスト桁数
	int screen_h {};		// テキスト行数
	int cursor_x {};		// カーソル位置(桁)
	int cursor_y {};		// カーソル位置(行)
	int origin_px {};		// テキスト画面の左上 X 座標(ピクセル)
	int origin_py {};		// テキスト画面の左上 Y 座標(ピクセル)

	// NVRAM
	// 本来他デバイスはすべてグローバル参照できるのだが、LUNA の NVRAM は
	// RTC と兼用になっていて、RTCDevice *gRTC として見えるので
	// MK48T02Device にするには dynamic_cast する必要がある。
	MK48T02Device *nvram {};

 private:
	// キー入力
	void ROM_Keyin();

	bool is_shift {};			// SHIFT キー押し下げ中
	int mousecnt {};			// マウス入力をスキップするためのカウンタ

	// LUNA キーコード -> ASCII 変換テーブル
	static const uint8 lunakey2asciitable[128];
	static const uint8 lunakey2shifttable[128];
};
