//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// SRAM
//

// SRAM はビッグエンディアンのままメモリに置かれるので、
// 必要ならアクセス時に都度変換すること。

#include "sram.h"
#include "config.h"
#include "mainapp.h"
#include "ram.h"
#include <array>

// グローバル参照用
SRAMDevice *gSRAM;

// コンストラクタ
SRAMDevice::SRAMDevice()
	: inherited("SRAM")
{
	devaddr = baseaddr;
	devlen  = 0x4000;
}

// デストラクタ
SRAMDevice::~SRAMDevice()
{
	gSRAM = NULL;
}

bool
SRAMDevice::Init()
{
	filename = gMainApp.GetVMDir() + "SRAM.DAT";
	file.SetFilename(filename);
	file.SetDispname(string_format("SRAM \"%s\"", filename.c_str()));
	mem = file.OpenCreate(16 * 1024);
	if (mem == NULL) {
		return false;
	}

	// SRAM の RAM 容量欄を設定値に同期させる。
	// この処理はアプリケーション起動時に一回行われるのみで、
	// その後の VM 内のリセット等では動作しない。
	bool sync = gConfig->Find("sram-sync-ramsize").AsInt();
	if (sync) {
		SyncRAMSize();
	}

	return true;
}

// リセット
void
SRAMDevice::ResetHard(bool poweron)
{
	// XXX 未調査
	writeable = false;
}

// SRAM の RAM 容量欄を設定値に同期させる。
void
SRAMDevice::SyncRAMSize()
{
	// SRAM の内容が無効なら何もしない。
	static const std::array<uint8, 7> magic {
		0x82, 0x77, '6', '8', '0', '0', '0'
	};
	for (int i = 0; i < magic.size(); i++) {
		if (mem[i] != magic[i]) {
			return;
		}
	}

	uint32 ramsize = gRAM->GetSize();

	// 値が違う時だけ表示したいので一旦読み出して比較
	uint32 cursize;
	cursize  = mem[0x08] << 24;
	cursize |= mem[0x09] << 16;
	cursize |= mem[0x0a] << 8;
	cursize |= mem[0x0b];
	if (cursize != ramsize) {
		// ここは Init() なので putmsg() は使えない
		warnx("SRAM: Update ramsize %dMB", ramsize / 1024 / 1024);

		mem[0x08] = ramsize >> 24;
		mem[0x09] = ramsize >> 16;
		mem[0x0a] = ramsize >> 8;
		mem[0x0b] = ramsize;
	}
}

// アドレスデコーダ
inline uint64
SRAMDevice::Decoder(uint32 addr) const
{
	uint32 offset = addr - baseaddr;
	assertmsg(offset < file.GetMemSize(),
		"offset=%x memlen=%x", offset, (int)file.GetMemSize());
	return offset;
}

uint64
SRAMDevice::Read8(uint32 addr)
{
	uint32 offset = Decoder(addr);
	uint32 data;

	data = Get8(offset);
	putlog(2, "$%06x.B -> $%02x", addr, data);

	return data;
}

uint64
SRAMDevice::Read16(uint32 addr)
{
	uint32 offset = Decoder(addr);
	uint32 data;

	data = Get16(offset);
	putlog(2, "$%06x.W -> $%04x", addr, data);

	return data;
}

uint64
SRAMDevice::Write8(uint32 addr, uint32 data)
{
	uint32 offset = Decoder(addr);

	if (writeable) {
		if (offset < 0x100) {
			putlog(1, "$%06x.B <- $%02x", addr, data);
		}
		mem[offset] = data;
	} else {
		putlog(1, "Write protected $%06x.B <- $%02x", addr, data);
	}
	return 0;
}

uint64
SRAMDevice::Write16(uint32 addr, uint32 data)
{
	uint32 offset = Decoder(addr);

	if (writeable) {
		if (offset < 0x100) {
			putlog(1, "$%06x.W <- $%04x", addr, data);
		}
		*(uint16 *)&mem[offset] = htobe16(data);
	} else {
		putlog(1, "Write protected $%06x.W <- $%04x", addr, data);
	}
	return 0;
}

uint64
SRAMDevice::Peek8(uint32 addr)
{
	uint32 offset = Decoder(addr);

	return Get8(offset);
}

void
SRAMDevice::WriteEnable(bool value)
{
	if (writeable == false && value == true) {
		putlog(1, "Write Enable");
	} else if (writeable == true && value == false) {
		putlog(1, "Write Disable");
	}

	writeable = value;
}

// SRAM から副作用なく読み出す。
// ただしホストファイル起動モードなら改変した値を返す。
uint32
SRAMDevice::Get8(uint32 offset) const
{
	if (gMainApp.exec_file) {
		switch (offset) {
		 case 0x0c:			// ROM 起動アドレス(.L)
			return plutoaddr >> 24;
		 case 0x0d:
			return (plutoaddr >> 16) & 0xff;
		 case 0x0e:
			return (plutoaddr >>  8) & 0xff;
		 case 0x0f:
			return plutoaddr & 0xff;

		 case 0x18:			// 起動デバイス(.W) を ROM($a000) にする
			return 0xa0;
		 case 0x19:
			return 0x00;

		 default:
			break;
		}
	}

	// 他は SRAM から
	return mem[offset];
}

// SRAM から副作用なくワードで読み出す。
// ただしホストファイル起動モードなら改変した値を返す。
uint32
SRAMDevice::Get16(uint32 offset) const
{
	uint32 data;

	data  = Get8(offset) << 8;
	data |= Get8(offset + 1);

	return data;
}

// SRAM から副作用なくロングワードで読み出す。
// ただしホストファイル起動モードなら改変した値を返す。
uint32
SRAMDevice::Get32(uint32 offset) const
{
	uint32 data;

	data  = Get16(offset) << 16;
	data |= Get16(offset + 2);

	return data;
}
