//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// LUNA88K のシステムコントローラのような何か
//

// IODevice
//    |
//    +-- InterruptDevice (割り込みコントローラの共通部)
//          |
//          +-- M680x0Interrupt (m68k 割り込みコントローラの共通部)
//          |     |
//          |     +-- X68030Interrupt  (X68030 の割り込みコントローラ)
//          |     +-- LunaInterrupt    (LUNA-I の割り込みコントローラ)
//          |
//          +-- PEDECDevice (X68030 の Lv1 担当コントローラ)
//          |
//          |  +---------------+
//          +--| SysCtlrDevice | (LUNA88K の割り込みコントローラ)
//             +---------------+

// この辺を全部一つのデバイスで受け付けておくか。BusIO はまた後で考える。
// $6500000n.b	HW 割り込みレベル (n*4 = CPU#n)
//   $65000000.b	CPU#0
//   $65000004.b	CPU#1
//   $65000008.b	CPU#2
//   $6500000c.b	CPU#3
// $6900000n	ソフトウェア割り込み(?)
// $6d00000n	CPU#n をリセット(?)
// $6d000010	全 CPU をリセット

// LUNA88K
// 割り込みコントロールレジスタ
// $6500_000n
//  n: CPU ID * 4 (0,4,8,c)
//  この16Mページ内はミラーされている模様。
//  下位2bitしかデコードしていない?
// 書き込み時:
//    3                   2                   1
//  1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
// +-+-+-+-+-+-+---------------------------------------------------+
// |6|5|4|3|2|1|                                                   |
// +-+-+-+-+-+-+---------------------------------------------------+
// bit31: 1=割り込み 6 イネーブル
// bit30: 1=割り込み 5 イネーブル
// bit29: 1=割り込み 4 イネーブル
// bit28: 1=割り込み 3 イネーブル
// bit27: 1=割り込み 2 イネーブル
// bit26: 1=割り込み 1 イネーブル
// m68k と異なり、個別の割り込みごとに許可/禁止できる。
// 該当ビットに 1 を書き込むと許可。
// 空欄部分は無視される模様。
//
// 読み込み時
//    3                   2                   1
//  1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
// +-----+---------+-+-+-+-+-+-+-----------------------------------+
// |Level|         |6|5|4|3|2|1|                                   |
// +-----+---------+-+-+-+-+-+-+-----------------------------------+
// bit31-29: 発生している最上位レベルの割り込み。範囲は7-0
// bit23-18: 書き込んだ割り込みマスク。
// 空欄部分は「不定」の模様。
//
// bit31-29 のレベルは、割り込みコントローラに入力
// されている割り込み信号線6本を、マスク処理した後、
// プライオリティバイナリエンコーダの値がそのまま出力されている模様。
// ラッチ機構は無いようだ。
// したがってマスクされている割り込み(マスクビットが 0) の
// 値が出力されることはない。
// 同時に入力があった場合は、数字の大きいほうが出力される。
// レベル7 はマスクなしでプライオリティエンコーダに入力されているようだ。
//
// 割り込み入力がないときは 0 になる模様。
// 本体 ABRT スイッチを押すと、押している間、すべての CPU の
// level が 111(=7) になる。離すと戻る。
// 88100 には NMI が無いので、通常の割り込みと同じように処理している？
// (なので IND 状態だとABRT できない。いいのかそれで)
// (OpenBSD6.7 の割り込みハンドラはABRTには対応していないように読める)
//
// 想定される回路
//
// [割り込みソース]
//  1:SOFTINT レジスタ $6900000n
//  |   2:(XP 低レベル?)
//  |   |   3:SPC
//  |   |   |   4:(LANCE|拡張スロット?)
//  |   |   |   |   5:(拡張PC98?|シリアル?)
//  |   |   |   |   |   6:(XP 高レベル?|システムクロック?)
//  |   |   |   |   |   |   7:ABRTスイッチ
//  V   V   V   V   V   V   |
// +---------------------+  |
// |   割り込みマスク    |  |  (R:bit23-18, W:bit31-26)
// +---------------------+  |
//  |   |   |   |   |   |   |
//  V   V   V   V   V   V   V
// +-------------------------+
// |プライオリティエンコーダ |-> (R:bit31-29)
// +-------------------------+
//          | | |
//          V V V
//        (Wired-OR)
//            |
//        [M88K INT]
//
// LUNA-I とはレベル配置が違う。
//
// 割り込みソース調査状況
// 1:SOFTINT レジスタ $6900000n
//   ROM のメモリコマンドで確認。OpenBSD machdep.c の記述の通り。
// 2:XP 低レベル
//   LUNA-I からの類推及び、OpenBSD mainbus.c に記述あり。OpenBSD のほうは
//   低レベルかどうかは不明。
// 3:SPC
//   ROM1.2, OpenBSD ともに使用。
// 4:LANCE
//   ROM はまだ不明。OpenBSD は使用。
// 4:拡張スロット
//   LUNA-I からの類推のみ。不明。
// 5:拡張PC98
//   LUNA-I からの類推のみ。LUNA88K にはスロットがないようなので、
//   LUNA88K2 が必要。
//   OpenBSD mainbus.c に記述あり。
// 5:シリアル(7201)
//   OpenBSD mainbus.c に記述。
// 6:XP 高レベル
//   LUNA-I からの類推。存在するかどうかも不明。
//   ただ、これをなくすと XP ファームウェアまで作り直しなので、
//   なくさないんじゃないかなあ。OMRON ファームで使ってなければ
//   さっくり廃止されているかもしれない(PSGPAM が使用...レベル1 に
//   変えればいいだけだが)
// 6:システムクロック
//   ROM のメモリコマンドではレベル6割り込みが常時上がっているように観測
//   できるので、おそらくシステムクロック割り込みが上がっている。
//   OpenBSD maibus.c にも記述。
// 7:ABRTスイッチ
//   ROM のメモリコマンドで確認。
//
//
// SOFTINT レジスタ $6900000n
//  $69000000: CPU0
//  $69000004: CPU1
//  $69000008: CPU2
//  $6900000c: CPU3
// 挙動:
// この32ビットのどこかにライトアクセスすると、INT1 をアサートする。
// この32ビットのどこかをリードアクセスすると、INT1 をネゲートする。
// データバスは無視される。
// 0x00 を書いても、アサートされることに注意。
// リードアクセスで読み込まれる値も不定のようだ。
// 要はアドレスバスとR/W信号線しか見ていない。

#include "sysctlr.h"
#include "ethernet.h"
#include "mpu88xx0.h"
#include "nmi.h"
#include "prom.h"
#include "sio.h"
#include "spc.h"
#include "sysclk.h"

// コンストラクタ
SysCtlrDevice::SysCtlrDevice()
	: inherited("SystemCtlr")
{
	AddAlias("SysCtlr");

	monitor.func = ToMonitorCallback(&SysCtlrDevice::MonitorUpdate);
	monitor.SetSize(25, 10);
	monitor.Regist(ID_MONITOR_INTERRUPT);
}

// デストラクタ
SysCtlrDevice::~SysCtlrDevice()
{
}

// リセット
void
SysCtlrDevice::ResetHard(bool poweron)
{
	// さすがにリセットくらいはされるだろう…
	inherited::ResetHard(poweron);

	for (int i = 0; i < 4; i++) {
		// XXX 初期値は分からないがとりあえず割り込み禁止にしておく
		cpu[i].maskval = 0xfc000000;
		cpu[i].intmask = MakeIntMask(cpu[i].maskval, i);
		cpu[i].ipl = 0;
	}
}

uint64
SysCtlrDevice::Read32(uint32 addr)
{
	int n = (addr >> 2) & 3;

	if (0x65000000 <= addr && addr < 0x65000010) {
		// bit31-29 に発生している最上位の割り込みレベル
		// bit23-18 に設定したレベル 6-1 の割り込みマスク
		uint32 data = (cpu[n].ipl << 29) | (cpu[n].maskval >> 8);
		putlog(2, "割り込み保持読み出し(CPU#%d) -> $%08x", n, data);
		return data;
	}

	if (0x69000000 <= addr && addr < 0x69000010) {
		// 読み出しで SOFTINT ネゲート (値は関係ない)
		putlog(1, "Softint(CPU#%d) Negate", n);
		intmap &= ~(IntmapSoft0 << n);
		ChangeInterrupt();
		return 0xffffffff;
	}

	if (0x6d000000 <= addr && addr < 0x6d000010) {
		putlog(0, "Read CPU%d reset (NOT IMPLEMENTED)", n);
		return 0xffffffff;
	}
	if (addr == 0x6d000010) {
		putlog(0, "Read All CPU reset (NOT IMPLEMENTED)");
		return 0xffffffff;
	}

	putlog(0, "Read $%08x (NOT IMPLEMENTED)", addr);
	return 0xffffffff;
}

uint64
SysCtlrDevice::Write32(uint32 addr, uint32 data)
{
	int n = (addr & 15) >> 2;

	if (0x65000000 <= addr && addr < 0x65000010) {
		data &= 0xfc000000;
		// 割り込みレベル変えるたびに表示されるのですごく多い。
		// さらに同レベルが書き込まれることも多い…。
		if (cpu[n].maskval != data) {
			putlog(2, "Interrupt Mask(CPU#%d) <- $%08x", n, data);
		} else {
			putlog(3, "Interrupt Mask(CPU#%d) <- $%08x", n, data);
		}

		// 設定値を保存
		cpu[n].maskval = data;
		// 内部マスクを書き換えて割り込み状態を更新
		cpu[n].intmask = MakeIntMask(data, n);
		ChangeInterrupt();
		return 0;
	}
	if (0x69000000 <= addr && addr < 0x69000010) {
		// 書き込みで SOFTINT アサート (値は関係ない)
		putlog(1, "Softint(CPU#%d) Assert 0x%x", n, data);
		counter[__builtin_clz(intmap)]++;
		intmap |= IntmapSoft0 << n;
		ChangeInterrupt();
		return 0;
	}
	if (0x6d000000 <= addr && addr < 0x6d000010) {
		putlog(0, "Write CPU%d reset (NOT IMPLEMENTED)", n);
		return 0;
	}
	if (addr == 0x6d000010) {
		putlog(1, "Reset All CPUs");
		// XXX このあたりもう少し整理する必要がある
		// 今は MPU.ResetHard() が func を ResetCallback に、time をリセット
		// 間隔に書き換えて Start() した後、この書き込み命令から戻ったところの
		// ExecCallback が time だけ所要サイクルに書き換えて再度 Start() する
		// ため、時間はともかく幸い ResetCallback() が呼ばれることで動いている。
		gPROM0->ResetHard(false);
		gMPU->ResetHard(false);
		ResetHard(false);
		return 0;
	}

	putlog(0, "Write $%08x <- $%08x (NOT IMPLEMENTED)", addr, data);
	return 0;
}

uint64
SysCtlrDevice::Peek8(uint32 addr)
{
	int n = (addr >> 2) & 3;
	uint32 data;

	if (0x65000000 <= addr && addr < 0x65000010) {
		data = (cpu[n].ipl << 29) | (cpu[n].maskval >> 8);
	} else {
		data = 0xffffffff;
	}
	switch (addr & 3) {
	 case 0:
		return data >> 24;
	 case 1:
		return (data >> 16) & 0xff;
	 case 2:
		return (data >>  8) & 0xff;
	 case 3:
		return data & 0xff;
	 default:
		break;
	}
	__unreachable();
}

void
SysCtlrDevice::MonitorUpdate(Monitor *, TextScreen& screen)
{
	int y;
	struct {
		const char *name;
		uint32 map;
	} table[] = {
		{ "NMI",	IntmapNMI },
		{ "Clock",	IntmapSysClk },
		{ "SIO",	IntmapSIO },
		{ "Lance",	IntmapLance },
		{ "SPC",	IntmapSPC },
		{ "Soft#0",	IntmapSoft0 },
		{ "Soft#1",	IntmapSoft1 },
		{ "Soft#2",	IntmapSoft2 },
		{ "Soft#3",	IntmapSoft3 },
	};

	// CPU の割り込み可否は本当は関係ないけど、便宜的に一緒に表示したい
	bool ei = gMPU88xx0->GetCPU()->IsIntrEnable();

	screen.Clear();
	y = 0;
	// XXX CPU の実装数によってウィンドウ幅を変える。
	// はみ出したところは描画されないのでヘッダは4つ分出力しておく
	screen.Print(0, y++, "Lv Device     Count  CPU0 CPU1 CPU2 CPU3");
	for (int i = 0; i < countof(table); i++) {
		uint32 map = table[i].map;
		int clz = __builtin_clz(map);
		int lv = (31 - clz) / 4;
		screen.Print(0, y, "%d", lv);
		screen.Print(2, y, TA::OnOff(intmap & map), "%s", table[i].name);
		screen.Print(8, y, "%11" PRIu64, counter[clz]);

		// CPU ごとのマスク
		// XXX 今は CPU は1つしかない
		for (int n = 0; n < 1; n++) {
			// Softint の他 CPU 分は
			// 動作としては常にマスクされてて通過しないから問題ないが、
			// 見せる分にはマスクされているではなく無効と見せたほうがいい。
			if (lv == 1 && map != (IntmapSoft0 << n)) {
				screen.Puts(21 + n * 5, y, "----");
			} else if ((cpu[n].intmask & map) == 0) {
				screen.Puts(21 + n * 5, y, "Mask");
			} else if (!ei) {
				screen.Puts(21 + n * 5, y, "IND");
			}
		}

		y++;
	}
}

// デバイスから Intmap 値を引く
uint32
SysCtlrDevice::GetIntmap(Device *source) const
{
	if (__predict_true(source == gSysClk)) {
		return IntmapSysClk;
	}
	if (__predict_true(source == gSIO)) {
		return IntmapSIO;
	}
	if (__predict_true(source == gEthernet)) {
		return IntmapLance;
	}
	if (__predict_true(source == gSPC)) {
		return IntmapSPC;
	}
	if (source == gNMI) {
		return IntmapNMI;
	}
	PANIC("Unknown interrupt source?");
}

// 割り込みアクノリッジはない
int
SysCtlrDevice::InterruptAcknowledge(int lv)
{
	return 0;
}

// 割り込み状態を更新
void
SysCtlrDevice::ChangeInterrupt()
{
	// XXX MPU 側がまだマルチプロセッサに対応してないので CPU#0 だけ
	for (int i = 0; i < 1; i++) {
		uint32 localmap = intmap & cpu[i].intmask;

		// intmap の最下位ビットは常に立ててあり localmap は 0 にならない
		int newipl = (uint)(31 - __builtin_clz(localmap)) / 4;

		// 割り込みレベルが変わったら、新しいレベルを通知
		if (newipl != cpu[i].ipl) {
			cpu[i].ipl = newipl;
			// MPU 側は 1 レベルしかない
			gMPU->Interrupt(cpu[i].ipl ? 1 : 0);
		}
	}
}

// 割り込みマスク設定レジスタの書き込み値から内部用のマスクを作成
// data は bit31-26 のみ使用。他のビットは不問。
// n は CPU# (Softint 用)。
/*static*/ uint32
SysCtlrDevice::MakeIntMask(uint32 data, int n)
{
	uint32 val = 0xf0000001;	// Lv7 は NMI、最下位は番兵用

	uint32 mask = 0x0f000000;
	for (int i = 6; i >= 1; i--) {
		if ((int32)data < 0) {
			val |= mask;
		}
		mask >>= 4;
		data <<= 1;
	}

	// Softint は他 CPU 分のビットを下ろす
	val &= (0xffffff0f | (IntmapSoft0 << n));

	return val;
}
