//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// システムポート
//

#include "sysport.h"
#include "keyboard.h"
#include "mpu.h"
#include "nmi.h"
#include "power.h"
#include "ram.h"
#include "romimg_x68k.h"
#include "sram.h"
#include <array>

// グローバル参照用
SysportDevice *gSysport;

// コンストラクタ
SysportDevice::SysportDevice()
	: inherited("SystemPort")
{
	AddAlias("SysPort");

	devaddr = baseaddr;
	devlen  = 0x2000;
}

// デストラクタ
SysportDevice::~SysportDevice()
{
	gSysport = NULL;
}

// リセット
void
SysportDevice::ResetHard(bool poweron)
{
	sysport.contrast = 0;	// XXX 未調査
	sysport.ram_wait = 0;
	sysport.rom_wait = 0;
	sysport.key_ctrl = false;
	sysport.pwoff_count = 0;
	gPower->SetSystemPowerOn(true);
}

uint64
SysportDevice::Read(uint32 offset)
{
	uint32 data;

	switch (offset) {
	 case SYSPORT::CONTRAST:
		data = 0xf0 | sysport.contrast;
		break;
	 case SYSPORT::SCOPE3D:
		data = 0xff;
		break;
	 case SYSPORT::IMAGEUNIT:
		data = 0xff;
		break;
	 case SYSPORT::KEY:
		// XXX bit3 以外は要実機確認
		data = 0xf7;
		if (gKeyboard->IsConnected()) {
			data |= 0x08;
		}
		break;
	 case SYSPORT::WAIT:
		data = 0xff;
		break;
	 case SYSPORT::MPU:
		data = 0xdc;
		break;
	 case SYSPORT::SRAMWP:
		data = 0xff;
		break;
	 case SYSPORT::POWEROFF:
		data = 0xff;
		break;
	 default:
		__unreachable();
	}
	putlog(2, "$%06x -> $%02x", gMPU->GetPaddr(), data);
	return data;
}

uint64
SysportDevice::Write(uint32 offset, uint32 data)
{
	switch (offset) {
	 case SYSPORT::CONTRAST:
		sysport.contrast = data & 15;
		break;

	 case SYSPORT::SCOPE3D:
	 case SYSPORT::IMAGEUNIT:
		break;

	 case SYSPORT::KEY:
		data &= 0x0e;
		// bit1
		if ((data & 0x02)) {
			putlog(0, "$e8e007 <- $%02x (HRL NOT IMPLEMENTED)", data);
		}
		// bit2: NMI リセット
		if ((data & 0x04)) {
			gNMI->NegateNMI();
		}
		sysport.key_ctrl = (data & 0x08);
		break;

	 case SYSPORT::WAIT:
	 {
		// 設定値で保持
		sysport.rom_wait = data >> 4;
		sysport.ram_wait = data;

		// ROM に指示。常に設定値 + 2 でよい
		gIPLROM1->SetWait(sysport.rom_wait + 2);
		gIPLROM2->SetWait(sysport.rom_wait + 2);
		gCGROM->SetWait(sysport.rom_wait + 2);
		// RAM に指示。InsideOut p.124
		int wait = sysport.ram_wait;
		if (wait > 0)
			wait += 2;
		gRAM->SetWait(wait);
		break;
	 }

	 case SYSPORT::MPU:
		break;

	 case SYSPORT::SRAMWP:
		if (data == 0x31) {
			gSRAM->WriteEnable(true);
		} else {
			gSRAM->WriteEnable(false);
		}
		break;

	 case SYSPORT::POWEROFF:
	 {
		static const std::array<uint8, 3> sig { 0x00, 0x0f, 0x0f };

		if (sysport.pwoff_count >= sig.size()) {
			// すでにカウントオーバーしている場合何もしない?
			putlog(2, "$e8e00f (POWEROFF) <- %02x", data);
			break;
		}

		if (data == sig[sysport.pwoff_count]) {
			sysport.pwoff_count++;
		} else {
			sysport.pwoff_count = 0;
		}
		putlog(2, "$e8e00f (POWEROFF) <- %02x (count=%d)",
			data, sysport.pwoff_count);
		if (sysport.pwoff_count == sig.size()) {
			// pwoff_count はここではクリアせず、電源オン時に再初期化する。
			putlog(1, "$e8e00f (POWEROFF) Power Off");
			gPower->SetSystemPowerOn(false);
		}
		break;
	 }

	 default:
		__unreachable();
	}
	return 0;
}

uint64
SysportDevice::Peek(uint32 offset)
{
	uint32 data;

	switch (offset) {
	 case SYSPORT::CONTRAST:
		data = 0xf0 | sysport.contrast;
		break;
	 case SYSPORT::SCOPE3D:
		data = 0xff;
		break;
	 case SYSPORT::IMAGEUNIT:
		data = 0xff;
		break;
	 case SYSPORT::KEY:
		// XXX bit3 以外は要実機確認
		data = 0xf7;
		if (gKeyboard->IsConnected()) {
			data |= 0x08;
		}
		break;
	 case SYSPORT::WAIT:
		data = 0xff;
		break;
	 case SYSPORT::MPU:
		data = 0xdc;
		break;
	 case SYSPORT::SRAMWP:
		data = 0xff;
		break;
	 case SYSPORT::POWEROFF:
		data = 0xff;
		break;
	 default:
		__unreachable();
	}
	return data;
}
