//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// TVRAM
//

// TVRAM (PlaneVRAM) はロングワード単位でホストバイトオーダ配置なので、
// バイトアクセス時は HLB()、ワードアクセス時は HLW() マクロを使用のこと。

#include "tvram.h"
#include "mpu.h"
#include "videoctlr.h"

// コンストラクタ
TVRAMDevice::TVRAMDevice()
	: inherited("TVRAM", 1024, 1024)
{
	devaddr = baseaddr;
	devlen  = 512 * 1024;

	// プレーン数は4枚固定
	nplane = NPLANE;

	modify_mask = 0xff00;

	mem.reset(new uint8[devlen]);
}

// デストラクタ
TVRAMDevice::~TVRAMDevice()
{
}

bool
TVRAMDevice::Init()
{
	// 加工済みパレットを取得
	palette = gVideoCtlr->GetHostPalette();

	return true;
}

// リセット
void
TVRAMDevice::ResetHard(bool poweron)
{
	// XXX ここ?
	Invalidate();
}

inline uint64
TVRAMDevice::Decoder(uint32 addr) const
{
	return addr - baseaddr;
}

uint64
TVRAMDevice::Read8(uint32 addr)
{
	gMPU->AddCycle(9);	// Inside/Out p.135
	uint32 offset = Decoder(addr);
	return mem[HLB(offset)];
}

uint64
TVRAMDevice::Read16(uint32 addr)
{
	gMPU->AddCycle(9);	// Inside/Out p.135
	uint32 offset = Decoder(addr);
	return *(uint16 *)&mem[HLW(offset)];
}

uint64
TVRAMDevice::Write8(uint32 addr, uint32 data)
{
	uint32 offset = Decoder(addr);
	uint8 *ptr;
	uint8 mask8;

	gMPU->AddCycle(9);	// Inside/Out p.135

	// XXX とりあえずね
	putlog(3, "Write $%06x.B <- $%02x", addr, data);

	if (tvram.men) {
		if ((addr & 1) == 0) {
			mask8 = tvram.mask >> 8;
		} else {
			mask8 = tvram.mask;
		}
	} else {
		mask8 = 0;
	}

	if (tvram.sa) {
		// 同時アクセス
		uint32 poffset = offset & 0x1ffff;
		for (int plane = 0; plane < 4; plane++) {
			if (tvram.ap[plane]) {
				ptr = &mem[HLB(0x20000 * plane + poffset)];
				*ptr = (data & ~mask8) | (*ptr & mask8);
			}
		}
	} else {
		// 通常アクセス
		ptr = &mem[HLB(offset)];
		*ptr = (data & ~mask8) | (*ptr & mask8);
	}
	SetDirty(offset);
	return 0;
}

uint64
TVRAMDevice::Write16(uint32 addr, uint32 data)
{
	uint32 offset = Decoder(addr);
	uint16 *ptr;
	uint16 mask16;

	gMPU->AddCycle(9);	// Inside/Out p.135

	// XXX とりあえずね
	putlog(3, "Write $%06x.W <- $%04x", addr, data);

	if (tvram.men) {
		mask16 = tvram.mask;
	} else {
		mask16 = 0;
	}

	if (tvram.sa) {
		// 同時アクセス
		uint32 poffset = offset & 0x1ffff;
		for (int plane = 0; plane < 4; plane++) {
			if (tvram.ap[plane]) {
				ptr = (uint16 *)&mem[HLW(0x20000 * plane + poffset)];
				*ptr = (data & ~mask16) | (*ptr & mask16);
			}
		}
	} else {
		// 通常アクセス
		ptr = (uint16 *)&mem[HLW(offset)];
		*ptr = (data & ~mask16) | (*ptr & mask16);
	}
	SetDirty(offset);
	return 0;
}

uint64
TVRAMDevice::Peek8(uint32 addr)
{
	uint32 offset = Decoder(addr);
	return mem[HLB(offset)];
}

// offset の位置に対応する更新フラグを立てる。
// offset は TVRAM 先頭からのバイトオフセット (プレーン先頭からではない)。
inline void
TVRAMDevice::SetDirty(uint32 offset)
{
	// VRAM は横 1024 ドット(128バイト) x 縦 1024 ドットで、これを
	// 横 128 ドット(16バイト) x 縦 16 ドットずつのブロックに分割する。
	// ブロック数は横 8個 x 縦 64個。
	//
	//              1                               0
	//        3 2 1 0 f e d c b a 9 8 7 6 5 4 3 2 1 0
	//       +-----------+-------+-----+-------+-----+
	// VRAM  |  Y座標 (1024dot)  | X (128byte) : 8bit|
	//       +-----------+-------+-----+-------+-----+
	//       :           :       :     :             :
	//       +-----------+-------+-----+-------------+
	// Block | Y(64blks) | 16dot | X(8)|   128 dot   | *: X は 8 blocks
	//       +-----------+-------+-----+-------------+
	//        ＼          ＼     |     |
	//          ＼          ＼   |     |
	//            ＼          ＼ |     |
	//               +-----------+-----+
	// Dirty         | Y(64blks) | X(8)|
	//               +-----------+-----+
	//                8 7 6 5 4 3 2 1 0

	// x 座標(バイト単位 7 bit)の上位 3 bit
	int x = (offset >> 4) & 0x07;
	// y 座標(10bit) の上位 6 bit
	int y = (offset >> 8) & 0x1f8;

	// からインデックスを求める
	int b = x + y;

	dirty[b] = 1;

#if defined(PERF_HIT)
	// どのロングワードが更新されたか
	dirty_word[(offset & 0x1ffff) >> 2] = true;
#endif
}

void
TVRAMDevice::SetScrollX(int x)
{
	xscroll = x;
}

void
TVRAMDevice::SetScrollY(int y)
{
	yscroll = y;
	gRenderer->Invalidate2();
}

// CRTC R21 での同時アクセス設定を反映する。
// CRTC から呼ばれる。
void
TVRAMDevice::SetAccessPlane(uint16 data)
{
	tvram.cp[0] = (data & 0x01);
	tvram.cp[1] = (data & 0x02);
	tvram.cp[2] = (data & 0x04);
	tvram.cp[3] = (data & 0x08);
	tvram.ap[0] = (data & 0x10);
	tvram.ap[1] = (data & 0x20);
	tvram.ap[2] = (data & 0x40);
	tvram.ap[3] = (data & 0x80);
	tvram.sa = (data & 0x100);
	tvram.men = (data & 0x200);

	if (0) {
		if (tvram.sa) {
			putlog(2, "テキスト画面同時アクセス有効 %c%c%c%c",
				tvram.ap[0] ? '0' : '-',
				tvram.ap[1] ? '1' : '-',
				tvram.ap[2] ? '2' : '-',
				tvram.ap[3] ? '3' : '-');
		} else {
			putlog(2, "テキスト画面同時アクセス無効");
		}
	}
}

// CRTC R23 でのアクセスマスク設定を反映する。
// CRTC から呼ばれる。
void
TVRAMDevice::SetAccessMask(uint16 data)
{
	tvram.mask = data;
}

// ラスターコピーを開始する。
// src_ras, dst_ras はライン番号 (レジスタ設定値を4倍したもの)。
void
TVRAMDevice::RasterCopy(int src_ras, int dst_ras)
{
	for (int plane = 0; plane < 4; plane++) {
		if (tvram.cp[plane]) {
			uint8 *src = &mem[0x20000 * plane + src_ras * 128];
			uint8 *dst = &mem[0x20000 * plane + dst_ras * 128];
			memcpy(dst, src, 128 * 4);
		}
	}

	// dst からの4ラインを全部 dirty にする。
#if !defined(PERF_HIT)
	// dirty は、1ブロックが縦 16dot 収容しているので、4の倍数から始まる
	// 4ラインは必ず1ブロックに収まる。横方向は1ブロックにつき 128dot 収容して
	// いるので全部で8ブロック。
	// よって dirty を立てるだけなら 8回の呼び出しで済む。
	for (int x = 0; x < 8; x++) {
		SetDirty(dst_ras * 128 + (x << 4));
	}
#else
	// PERF_HIT の更新を行う場合は律儀に全プレーンの全ロングワード分ループする
	for (int plane = 0; plane < 4; plane++) {
		for (int i = 0; i < 4; i++) {
			int y = dst_ras + i;
			for (int x = 0; x < 128 / sizeof(uint32); x++) {
				SetDirty(0x20000 * plane + y * 128 + x * sizeof(uint32));
			}
		}
	}
#endif
}

// (x, y) 位置のピクセル情報を screen に描画。(ビットマップモニタの情報欄)
void
TVRAMDevice::UpdateInfo(TextScreen& screen, int x, int y) const
{
	screen.Clear();

	//0         1         2         3         4         5         6
	//012345678901234567890123456789012345678901234567890123456789012345
	//X=1024 Y=2024: (P0) $e01234: $00 (%00000000) ColorCode=$d

	screen.Puts(0, 0, "X=     Y=");
	for (int i = 0; i < nplane; i++) {
		screen.Print(15, i, "(P%d)", i);
	}

	// カーソルが範囲外ならここまで。
	if (x < 0) {
		return;
	}

	screen.Print(2, 0, "%4d Y=%4d", x, y);

	// 1プレーン内のバイトオフセット
	uint32 planeoffset = ((uint)y * 128) + ((uint)x / 8);

	uint cc = 0;
	for (int i = 0; i < nplane; i++) {
		// プレーン0先頭からのバイトオフセット
		uint32 totaloffset = 0x20000 * i + planeoffset;

		uint8 data = mem[HLB(totaloffset)];
		screen.Print(15, i, "(P%d) $%06x: $%02x (%%",
			i, baseaddr + totaloffset, data);

		// ビットパターン
		for (int m = 0x80; m != 0; m >>= 1) {
			screen.Putc((data & m) ? '1' : '0');
		}
		screen.Putc(')');

		// カーソル位置をハイライト
		screen.Locate(35 + (x % 8), i);
		screen.SetAttr(TA::Em);

		// カラーコードを計算
		if ((data & (0x80U >> (x % 8))) != 0) {
			cc |= (1U << i);
		}
	}

	screen.Print(45, 0, "ColorCode=$%x", cc);
}
