//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// メモリダンプウィンドウ
//

#include "wxdumpmonitor.h"
#include "wxmonitor.h"
#include "bus.h"
#include "debugger.h"

enum {
	ID_TEXT = IDGROUP_MEMDUMP,
	ID_PREV,
	ID_NEXT,

	ID_local_end,	// 最後に置く (チェック用)
};
static_assert(ID_local_end - 1 <= (int)IDGROUP_MEMDUMP_END, "ID exceeded");

// イベントテーブル
wxBEGIN_EVENT_TABLE(WXMemdumpWindow, inherited)
	EVT_TEXT_ENTER(ID_TEXT, WXMemdumpWindow::OnTextEnter)
	EVT_BUTTON(ID_PREV, WXMemdumpWindow::OnPrev)
	EVT_BUTTON(ID_NEXT, WXMemdumpWindow::OnNext)
wxEND_EVENT_TABLE()

// コンストラクタ
// (タイトルは UpdateAddr() でセットする)
WXMemdumpWindow::WXMemdumpWindow(wxWindow *parent, int idx_)
	: inherited(parent, wxID_ANY, wxEmptyString,
		DEFAULT_STYLE | wxRESIZE_BORDER)
{
	idx = idx_;

	// ↓+-------------------------+
	//   | 上枠 (アドレス、△、▽) |
	//   +-------------------------+
	//   | TextScreen              |
	//   +-------------------------+

	auto *topsizer = new wxBoxSizer(wxVERTICAL);

	// 上枠用の下敷きパネル
	auto *ctrlpanel = new WXBitmapPanel(this);
	topsizer->Add(ctrlpanel, 0, wxEXPAND);

	// 上枠用の横 sizer
	ctrlbox = new wxBoxSizer(wxHORIZONTAL);

	// アドレスコントロール
	addrctrl = new wxTextCtrl(this, ID_TEXT,
		wxEmptyString, wxDefaultPosition, wxDefaultSize,
		wxTE_PROCESS_ENTER);
	ctrlbox->Add(addrctrl);
	ctrlbox->AddSpacer(10);

	// △ボタン
	// XXX とりあえずね
	prev_btn = new wxButton(this, ID_PREV, wxT("△"));
	next_btn = new wxButton(this, ID_NEXT, wxT("▽"));
	ctrlbox->Add(prev_btn);
	ctrlbox->Add(next_btn);

	// 上枠の sizer と下敷きを紐付ける
	ctrlpanel->SetSizer(ctrlbox);
	ctrlbox->SetSizeHints(ctrlpanel);

	screen = new WXMonitorPanel(this, debugger_memdump_monitor(idx));
	topsizer->Add(screen, 1, wxEXPAND);

	SetSizer(topsizer);

	// ウィンドウサイズを確定させる
	fixed_height = ctrlbox->GetSize().y;
	DoSize();

	// 表示を初期化する
	screen->SetUserData(sticky_addr[idx]);
	UpdateAddr();
}

// デストラクタ
WXMemdumpWindow::~WXMemdumpWindow()
{
}

bool
WXMemdumpWindow::Layout()
{
	// 縦リサイズ可能レイアウト
	return LayoutTextVResize(screen, fixed_height);
}

// テキスト入力イベント
void
WXMemdumpWindow::OnTextEnter(wxCommandEvent& event)
{
	// 入力文字列(16進文字列)を数値に変換
	const wxString& wstr = event.GetString();
	uint32 val = strtoul((const char *)wstr.mb_str(), NULL, 16);

	// 16 に整列
	val &= ~15;

	// アドレスを更新
	union64 udata;
	udata.q = screen->GetUserData();
	udata.l = val;
	screen->SetUserData(udata.q);
	UpdateAddr();
}

// △(PREV) ボタン押下イベント
void
WXMemdumpWindow::OnPrev(wxCommandEvent& event)
{
	union64 udata;

	// 1ページ戻る (横16バイト固定 × 表示行数分)
	udata.q = screen->GetUserData();
	udata.l -= screen->GetRow() * 16;

	screen->SetUserData(udata.q);
	UpdateAddr();
}

// ▽(NEXT) ボタン押下イベント
void
WXMemdumpWindow::OnNext(wxCommandEvent& event)
{
	union64 udata;

	// 1ページ進む (横16バイト固定 × 表示行数分)
	udata.q = screen->GetUserData();
	udata.l += screen->GetRow() * 16;

	screen->SetUserData(udata.q);
	UpdateAddr();
}

// アドレス欄の表示を更新する
void
WXMemdumpWindow::UpdateAddr()
{
	uint32 addr = screen->GetUserData();

	// 二次情報にコピー。
	sticky_addr[idx] = addr;

	// EVT_TEXT を起こさず変更する
	addrctrl->ChangeValue(wxString(string_format("%08x", addr)));

	// カーソルを末尾に
	addrctrl->SetInsertionPointEnd();

	// タイトルを変更
	SetTitle(_("Memory Dump") + string_format(" %d: $%08x", idx, addr));
}

// ウィンドウ ID に対応する sticky address を返す
uint32
WXMemdumpWindow::GetStickyAddr(int id)
{
	int n = id - ID_MONITOR_MEMDUMP0;
	assert(n < sticky_addr.size());
	return sticky_addr[n];
}

// ウィンドウ ID に対応する sticky address を設定する
void
WXMemdumpWindow::SetStickyAddr(int id, uint32 addr)
{
	int n = id - ID_MONITOR_MEMDUMP0;
	assert(n < sticky_addr.size());
	sticky_addr[n] = addr;
}

// アドレス
/*static*/ std::array<uint32, MAX_MEMDUMP_MONITOR>
WXMemdumpWindow::sticky_addr;
