//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// LUNA LCD パネル
//

#include "wxlcdwindow.h"
#include "wxcolor.h"
#include "wxuimessage.h"
#include "lcd.h"

// パネルの大きさ (定数)
// N はとりあえずの倍率。
#define N (2)
#define PANEL_WIDTH  ((N) * (5 + 1) * 16 + 1)
#define PANEL_HEIGHT ((N) * (8 + 1) *  2 + 1)

//
// LUNA LCD パネル
//

// コンストラクタ
WXLCDPanel::WXLCDPanel(wxWindow *parent)
	: inherited(parent, wxSize(PANEL_WIDTH, PANEL_HEIGHT))
{
	ddram = gLCD->GetDDRAM();

	// VM からの通知を受け取る
	WXUIMessage::Connect(UIMessage::LCD, this,
		wxCommandEventHandler(WXLCDPanel::OnLCDChanged));

	SetBitmapBGColor(UD_CREAM);
}

// デストラクタ
WXLCDPanel::~WXLCDPanel()
{
	WXUIMessage::Disconnect(UIMessage::LCD,
		wxCommandEventHandler(WXLCDPanel::OnLCDChanged));
}

// LCD 状態変更イベント (UIMessage)
void
WXLCDPanel::OnLCDChanged(wxCommandEvent& event)
{
	Refresh();
}

// バックグラウンドバッファに描画
void
WXLCDPanel::Draw()
{
	for (int y = 0; y < 2; y++) {
		for (int x = 0; x < 16; x++) {
			uint8 ch = ddram[0x40 * y + x];
			DrawChar(x, y, ch);
		}
	}
}

// 1文字描画
void
WXLCDPanel::DrawChar(int x, int y, uint8 ch)
{
	const Color pal[2] = { UD_YELLOW_GREEN, UD_BROWN };
	const uint8 *font;

	font = gLCD->GetFont(ch);
	BitmapI1 src(font, 5, 8);

	bitmap.DrawScaleUp((1 + x * 6) * N, (1 + y * 9) * N, N, src, pal);
}


//
// LCD ウィンドウ
//

// コンストラクタ
WXLCDWindow::WXLCDWindow(wxWindow *parent)
	: inherited(parent, wxID_ANY, _("Front LCD"))
{
	panel = new WXLCDPanel(this);

	DoSize();
}

// デストラクタ
WXLCDWindow::~WXLCDWindow()
{
}


//
// LCD キャラクタパネル
//

#define PAD (1)
#define CP_WIDTH  (N * (PAD + 6 * (16 + 1 + 16) + PAD))
#define CP_HEIGHT (N * (PAD + 9 * 8 + PAD))

// コンストラクタ
WXLCDCharPanel::WXLCDCharPanel(wxWindow *parent)
	: inherited(parent, wxSize(CP_WIDTH, CP_HEIGHT))
{
}

// デストラクタ
WXLCDCharPanel::~WXLCDCharPanel()
{
}

// バックグラウンドバッファに描画
void
WXLCDCharPanel::Draw()
{
	const Color pal[2] = { UD_WHITE, UD_BLACK };

	// フォントは 5x8 (最下行はカーソル用に空けてあるので実際は 5x7)。
	// 実際には 0xe0 以降には 5x10 フォントが混在しているのだが、その
	// モードはサポートしていないので今の所全部 5x8 としてある。

	// cx, cy はこの文字の左上座標 (スケール前)
	int cx = PAD;
	int cy = PAD;
	for (int ch = 0; ch < 256; ch++) {
		const uint8 *font = gLCD->GetFont(ch);
		BitmapI1 src(font, 5, 8);
		bitmap.DrawScaleUp(cx * N, cy * N, N, src, pal);

		cx += 6;
		if (ch % 16 == 15) {
			// 行折り返し
			cx -= 6 * 16;
			cy += 9;
			// 0x80 以降は右側に作る
			if (ch == 127) {
				cx += 6 * (16 + 1/*中央の余白*/);
				cy = 1;
			}
		}
	}
}


//
// LCD モニタウィンドウ
//

// コンストラクタ
WXLCDMonitor::WXLCDMonitor(wxWindow *parent)
	: inherited(parent, wxID_ANY, _("LCD"))
{
	// +------------+
	// | TextScreen |
	// +------------+
	// | CharPanel  |
	// +------------+
	auto *topsizer = new wxBoxSizer(wxVERTICAL);

	// 上段、テキストスクリーン
	monpanel = new WXMonitorPanel(this, gMonitorManager->Get(ID_MONITOR_LCD));
	// 下辺は CharPanel の上余白が隣接しているため、自分のほうは付けない。
	monpanel->SetPadding(-1, -1, -1, 0);
	topsizer->Add(monpanel, 0, wxEXPAND);

	// 下段、キャラクタパネル
	chrpanel = new WXLCDCharPanel(this);
	topsizer->Add(chrpanel, 0, wxEXPAND);

	SetSizer(topsizer);

	// 大きさをそれに固定
	SetClientSize(GetSizer()->GetMinSize());
}

// デストラクタ
WXLCDMonitor::~WXLCDMonitor()
{
}
