//
// nono
// Copyright (C) 2022 nono project
// Licensed under nono-license.txt
//

//
// ログレベル設定ウィンドウ
//

#include "wxlogsetting.h"
#include "wxcolor.h"
#include "wxtextscreen.h"
#include "mainapp.h"
#include <algorithm>

//
// パネル部分
//

// イベントテーブル
wxBEGIN_EVENT_TABLE(WXLogSettingPanel, inherited)
	EVT_LEFT_DOWN(WXLogSettingPanel::OnLeftDown)
	EVT_LEFT_UP(WXLogSettingPanel::OnLeftUp)
wxEND_EVENT_TABLE()

// コンストラクタ
WXLogSettingPanel::WXLogSettingPanel(wxWindow *parent)
	: inherited(parent)
{
	// エイリアスを持つオブジェクトだけ抜き出す
	for (const auto obj : gMainApp.GetObjects()) {
		if (obj->GetAliases().empty() == false) {
			objs.emplace_back(obj);
		}
	}

	// ログ名でソート。
	// ただし普通にソートすると "(BusErr)" とかの括弧付きがアルファベット
	// より上に来てしまうが、ここではこれらは重要ではないので下にしたい。
	std::sort(objs.begin(), objs.end(), [](Object *a, Object *b) {
		std::string aname = a->GetName();
		std::string bname = b->GetName();
		if (aname[0] == '(') {
			aname[0] = '{';
		}
		if (bname[0] == '(') {
			bname[0] = '{';
		}
		return aname < bname;
	});

	// 名前の最大長
	maxnamelen = 0;
	for (const auto obj : objs) {
		maxnamelen = std::max(maxnamelen, (int)obj->GetName().length());
	}

	FontChanged();
}

// デストラクタ
WXLogSettingPanel::~WXLogSettingPanel()
{
}

// 桁数 col の左端ピクセル座標を返す
inline int
WXLogSettingPanel::Col2PX(int col) const
{
	return col * GetFontWidth() + Padding;
}

// 行数 row の上端ピクセル座標を返す
inline int
WXLogSettingPanel::Row2PY(int row) const
{
	// 縦はボタン描画の分少し広げてある
	return row * (GetFontHeight() + 3) + Padding;
}

// ピクセル x が含まれる桁を返す。
// py がパディング領域を指していると(テキストとしての)範囲外の値を返す。
inline int
WXLogSettingPanel::PX2Col(int px) const
{
	return (px - Padding) / GetFontWidth();
}

// ピクセル y が含まれる行を返す。
// py がパディング領域を指していると(テキストとしての)範囲外の値を返す。
inline int
WXLogSettingPanel::PY2Row(int py) const
{
	return (py - Padding) / (GetFontHeight() + 3);
}

// フォントサイズ変更
void
WXLogSettingPanel::FontChanged()
{
	inherited::FontChanged();

	// 行数×桁数を基本とする
	int col = maxnamelen + 10;
	int row = objs.size() + 1;

	int width  = Col2PX(col) + Padding;
	int height = Row2PY(row) + Padding;
	SetClientSize(width, height);
	SetMinClientSize(wxSize(width, height));
}

// ボタンのようなものを描画する。rect は外寸。
void
WXLogSettingPanel::DrawButton(const Rect& rect, bool button_pressed)
{
	int l = rect.x;
	int t = rect.y;
	int r = rect.GetRight();
	int b = rect.GetBottom();

	if (__predict_true(button_pressed == false)) {
		// この場合外寸 7x7、内寸 3x3。
		// 空白「　」のところが内側。
		// 「・」は背景色で塗りつぶす (オンオフ逆側の陰影があるため)
		//
		// □□□□□□■ <- (1)
		// □・・・・灰■ <- (2)
		// □・　　　灰■
		// □・　　　灰■
		// □・　　　灰■
		// □灰灰灰灰灰■ <- (3)
		// ■■■■■■■ <- (4)
		// ^ ^       ^ ^
		//(5)(6)    (7)(8)

		bitmap.DrawLineH(UD_WHITE, l, t, r);				// (1)
		bitmap.DrawLineH(BGPANEL,  l + 1, t + 1, r -1);		// (2)
		bitmap.DrawLineH(UD_GREY,  l + 1, b - 1, r);		// (3)
		bitmap.DrawLineH(UD_BLACK, l, b, r + 1);			// (4)
		bitmap.DrawLineV(UD_WHITE, l, t + 1, b);			// (5)
		bitmap.DrawLineV(BGPANEL,  l + 1, t + 2, b - 1);	// (6)
		bitmap.DrawLineV(UD_GREY,  r - 1, t + 1, b - 1);	// (7)
		bitmap.DrawLineV(UD_BLACK, r, t, b);				// (8)
	} else {
		// ■■■■■■□ <- (1)
		// ■灰灰灰灰灰□ <- (2)
		// ■灰　　　・□
		// ■灰　　　・□
		// ■灰　　　・□
		// ■灰・・・・□ <- (3)
		// □□□□□□□ <- (4)
		// ^ ^       ^ ^
		//(5)(6)    (7)(8)
		bitmap.DrawLineH(UD_BLACK, l, t, r);				// (1)
		bitmap.DrawLineH(UD_GREY,  l + 1, t + 1, r);		// (2)
		bitmap.DrawLineH(BGPANEL,  l + 2, b - 1, r);		// (3)
		bitmap.DrawLineH(UD_WHITE, l, b, r + 1);			// (4)
		bitmap.DrawLineV(UD_BLACK, l, t + 1, b);			// (5)
		bitmap.DrawLineV(UD_GREY,  l + 1, t + 1, b);		// (6)
		bitmap.DrawLineV(BGPANEL,  r - 1, t + 2, b);		// (7)
		bitmap.DrawLineV(UD_WHITE, r, t, b);				// (8)
	}
}

// 描画本体
void
WXLogSettingPanel::Draw()
{
	const int width  = GetFontWidth();
	const int height = GetFontHeight();
	int row;
	int y;

	// ヘッダ
	y = Row2PY(0);
	DrawStringSJIS(Col2PX(0), y, "Name");
	DrawStringSJIS(Col2PX(maxnamelen + 1), y, "LogLevel");

	row = 1;
	for (const auto obj : objs) {
		Rect rect;
		Color c;

		y = Row2PY(row);

		DrawStringSJIS(Col2PX(0), y, obj->GetName().c_str());

		// 左ボタン (領域はテキスト2文字分)
		int x = Col2PX(maxnamelen + 1);
		if (obj->loglevel < 0) {
			c = UD_GREY;
		} else {
			c = GetTextForeColor();
		}
		DrawStringSJIS(c, x + (width / 2), y, "<");
		rect = Rect(x, y - 1, width * 2 + 2, height + 2);
		DrawButton(rect, (row == -pressed));

		// 値
		char lvstr[2];
		if (__predict_false(obj->loglevel < 0)) {
			lvstr[0] = '-';
		} else if (__predict_true(obj->loglevel < 10)) {
			lvstr[0] = obj->loglevel + '0';
		} else {
			lvstr[0] = '?';
		}
		lvstr[1] = '\0';
		DrawStringSJIS(Col2PX(maxnamelen + 5), y, lvstr);

		// 値の枠
		x = Col2PX(maxnamelen + 4);
		rect = Rect(x, y - 1, width * 3, height + 1);
		int l = rect.x;
		int t = rect.y;
		int r = rect.GetRight();
		int b = rect.GetBottom();
		bitmap.DrawLineH(UD_GREY,  l, t, r);
		bitmap.DrawLineH(UD_WHITE, l, b, r + 1);
		bitmap.DrawLineV(UD_GREY,  l, t + 1, b);
		bitmap.DrawLineV(UD_WHITE, r, t, b);

		// 右ボタン
		x = Col2PX(maxnamelen + 8);
		if (obj->loglevel >= 9) {
			c = UD_GREY;
		} else {
			c = GetTextForeColor();
		}
		DrawStringSJIS(c, x + (width / 2), y, ">");
		rect = Rect(x, y - 1, width * 2 + 2, height + 2);
		DrawButton(rect, (row == pressed));

		row++;
	}
}

// 左クリック (Down) イベント
void
WXLogSettingPanel::OnLeftDown(wxMouseEvent& event)
{
	const wxPoint& pos = event.GetPosition();
	int x = PX2Col(pos.x);
	int y = PY2Row(pos.y);

	// 上下のパディング領域とヘッダ行を先に除いておく
	// (左右は影響しないので放置でいい)
	if (y < 1 || y >= objs.size()) {
		return;
	}

	Object *obj = objs[y - 1];
	int level = obj->loglevel;

	// クリックされた場所がボタンの位置か
	if (maxnamelen + 1 <= x && x <= maxnamelen + 2) {
		if (level > -1) {
			level--;
		}
		// "<" ボタンは負数で覚えておく (0 はないので大丈夫)
		pressed = -y;
	} else if (maxnamelen + 8 <= x) {
		// 今の所上限は未定だけど、表示上の都合で一桁にしておくか
		if (level < 9) {
			level++;
		}
		// ">" ボタンは正数で覚えておく
		pressed = y;
	}

	if (level != obj->loglevel) {
		obj->SetLogLevel(level);
	}

	Refresh();
}

// 左クリック (Up) イベント
void
WXLogSettingPanel::OnLeftUp(wxMouseEvent& event)
{
	// クリックを上げた場所は問わない
	pressed = 0;
	Refresh();
}


//
// ウィンドウ
//

// コンストラクタ
WXLogSettingWindow::WXLogSettingWindow(wxWindow *parent)
	: inherited(parent, wxID_ANY, _("Log Level Setting"))
{
	panel = new WXLogSettingPanel(this);

	DoSize();
}

// デストラクタ
WXLogSettingWindow::~WXLogSettingWindow()
{
}
