//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// プレーン VRAM モニターウィンドウ
//

#include "wxplanemonitor.h"
#include "wxcolor.h"
#include "wxuimessage.h"
#include "bt454.h"
#include "planevram.h"
#include "videoctlr.h"

enum {
	ID_PLANE0 = IDGROUP_PLANEVRAM,
	ID_PLANE1,
	ID_PLANE2,
	ID_PLANE3,
	ID_PLANE4,
	ID_PLANE5,
	ID_PLANE6,
	ID_PLANE7,
	ID_PALETTE,
	ID_SCALE,

	ID_local_end,	// 最後に置く (チェック用)
};
static_assert(ID_local_end - 1 <= (int)IDGROUP_PLANEVRAM_END, "ID exceeded");

#define ID_PLANE(id)	(ID_PLANE0 + (id))

//
// プレーン VRAM ビットマップ
//

// イベントテーブル
wxBEGIN_EVENT_TABLE(WXPlaneVRAMBitmapPanel, inherited)
	EVT_MOUSE_EVENTS(WXPlaneVRAMBitmapPanel::OnMouse)
wxEND_EVENT_TABLE()

// コンストラクタ
WXPlaneVRAMBitmapPanel::WXPlaneVRAMBitmapPanel(wxWindow *parent)
	: inherited(parent)
{
	plane_width  = gPlaneVRAM->GetComposite().GetWidth();
	plane_height = gPlaneVRAM->GetComposite().GetHeight();

	// まず等倍でセット
	scale = 1;
	SetScale(1);

	cursor = wxDefaultPosition;

	// VM からの通知を受け取る
	WXUIMessage::Connect(UIMessage::PALETTE, this,
		wxCommandEventHandler(WXPlaneVRAMBitmapPanel::OnPaletteChanged));
}

// デストラクタ
WXPlaneVRAMBitmapPanel::~WXPlaneVRAMBitmapPanel()
{
	WXUIMessage::Disconnect(UIMessage::PALETTE,
		wxCommandEventHandler(WXPlaneVRAMBitmapPanel::OnPaletteChanged));
}

// XXX 実際にはこの辺かどこかで更新チェックをしないといけない。
// 今は表示領域を毎回描画しているので無駄。

// ビットマッププレーンの内容を描画する
void
WXPlaneVRAMBitmapPanel::Draw()
{
	assert(view.GetRight()  < virtual_width);
	assert(view.GetBottom() < virtual_height);

	const BitmapI8& src = gPlaneVRAM->GetComposite();

	if (scale > 1) {
		Rect dr(0, 0, bitmap.GetWidth(), bitmap.GetHeight());
		bitmap.DrawScale(dr, src, &pal[0], view.x, scale, 1, view.y, scale, 1);
	} else {
		bitmap.DrawBitmap(0, 0, src, &pal[0], view);
	}
}

// プレーン選択変更
void
WXPlaneVRAMBitmapPanel::EnablePlane(int plane, bool value)
{
	if (value) {
		planemask |= 1U << plane;
	} else {
		planemask &= ~(1U << plane);
	}

	// パレット生成
	GenPalette();
}

// パレット適用変更
void
WXPlaneVRAMBitmapPanel::EnablePalette(bool value)
{
	apply_palette = value;

	// パレット生成
	GenPalette();
}

// VM からのパレット変更通知
void
WXPlaneVRAMBitmapPanel::OnPaletteChanged(wxCommandEvent& event)
{
	// パレット生成
	GenPalette();
}

// パレット生成
void
WXPlaneVRAMBitmapPanel::GenPalette()
{
	if (apply_palette) {
		// パレットを適用する場合
		const ColorXBGR *vmpal;
		// 雑に X680x0/LUNA 対応
		if (gVideoCtlr) {
			vmpal = gVideoCtlr->GetHostPalette();
		} else {
			vmpal = gBT454->GetHostPalette();
		}
		const Color bg = vmpal[0].ToColor();
		for (int i = 0; i < pal.size(); i++) {
			if ((i & planemask)) {
				int n = i % 16;
				pal[i] = vmpal[n].ToColor();
			} else {
				pal[i] = bg;
			}
		}
	} else {
		// パレットを適用しない場合
		for (int i = 0; i < pal.size(); i++) {
			if ((i & planemask)) {
				pal[i] = Color(255, 255, 255);
			} else {
				pal[i] = Color(0, 0, 0);
			}
		}
	}
}

// 倍率を設定
void
WXPlaneVRAMBitmapPanel::SetScale(int scale_)
{
	int before = scale;

	scale = scale_;

	// 仮想画面の大きさを計算
	virtual_width = plane_width * scale;
	virtual_height = plane_height * scale;

	// ビューのスケール変換は前後のスケールがいるのでここでやる
	view.x = view.x * scale / before;
	view.y = view.y * scale / before;
	view.w = view.w * scale / before;
	view.h = view.h * scale / before;

	SetMaxClientSize(wxSize(virtual_width, virtual_height));
}

// スクロール位置変更通知
void
WXPlaneVRAMBitmapPanel::OnScroll(wxScrollEvent& event)
{
	int orient = event.GetOrientation();
	int pos = event.GetPosition();

	if (orient == wxHORIZONTAL) {
		view.x = pos;
	} else {
		view.y = pos;
	}
}

// マウスイベント
void
WXPlaneVRAMBitmapPanel::OnMouse(wxMouseEvent& event)
{
	auto window = dynamic_cast<WXPlaneVRAMWindow*>(GetGrandParent());
	assert(window);

	// pt はこのパネル左上からの位置
	const wxPoint& pt = event.GetPosition();

	// pt がパネル内の座標なのに Leaving() になることがある。
	// pt がパネル外の座標なのに Leaving() にならないことがある。
	if (event.Leaving() == false && GetClientRect().Contains(pt)) {
		// cursor は仮想画面上の座標
		cursor.x = (view.x + pt.x) / scale;
		cursor.y = (view.y + pt.y) / scale;
		assertmsg(cursor.x < plane_width,  "cursor.x=%d", cursor.x);
		assertmsg(cursor.y < plane_height, "cursor.y=%d", cursor.y);
	} else {
		cursor = wxDefaultPosition;
	}
}


//
// プレーン VRAM パネル
//

// イベントテーブル
wxBEGIN_EVENT_TABLE(WXPlaneVRAMPanel, inherited)
	EVT_SIZE(WXPlaneVRAMPanel::OnSize)
wxEND_EVENT_TABLE()

// コンストラクタ
WXPlaneVRAMPanel::WXPlaneVRAMPanel(wxWindow *parent)
	: inherited(parent)
{
	// 表示パネル
	viewctl = new WXPlaneVRAMBitmapPanel(this);

	// スクロールバーを自分で持つ (配置は Layout() で行う)
	vscroll = new WXScrollBar(this, wxID_ANY, wxSB_VERTICAL);
	hscroll = new WXScrollBar(this, wxID_ANY, wxSB_HORIZONTAL);

	// 右下の空き地
	corner = new WXBitmapPanel(this);

	// Sizer 使わず自前でレイアウトする。
	SetAutoLayout(true);

	// 最小値は適当
	int vscr_width  = vscroll->GetClientSize().x;
	int hscr_height = hscroll->GetClientSize().y;
	SetMinClientSize(wxSize(16 + vscr_width, 16 + hscr_height));
	SetMaxClientSize(wxSize(viewctl->virtual_width + vscr_width,
		viewctl->virtual_height + hscr_height));

	// スクロールバーからの通知イベントは表示パネルへ回す
	vscroll->Connect(NONO_EVT_SCROLL,
		wxScrollEventHandler(WXPlaneVRAMBitmapPanel::OnScroll), NULL, viewctl);
	hscroll->Connect(NONO_EVT_SCROLL,
		wxScrollEventHandler(WXPlaneVRAMBitmapPanel::OnScroll), NULL, viewctl);
}

// デストラクタ
WXPlaneVRAMPanel::~WXPlaneVRAMPanel()
{
}

bool
WXPlaneVRAMPanel::Layout()
{
	if (inherited::Layout() == false) {
		return false;
	}

	// コントロールを配置
	const wxSize size = GetClientSize();
	int vscr_width  = vscroll->GetClientSize().x;
	int hscr_height = hscroll->GetClientSize().y;
	int view_width  = size.x - vscr_width;
	int view_height = size.y - hscr_height;
	viewctl->SetSize(0, 0, view_width, view_height);
	vscroll->SetSize(view_width, 0, vscr_width, view_height);
	hscroll->SetSize(0, view_height, view_width, hscr_height);
	corner->SetSize(view_width, view_height, vscr_width, hscr_height);

	// XXX 内部のコントロールのサイズが確定してからでないと DoSize() できない
	DoSize(size);

	return true;
}

// サイズ変更イベント
void
WXPlaneVRAMPanel::OnSize(wxSizeEvent& event)
{
	// Mac ではウィンドウ作成時に 0 以下の値が来ることがある。
	// GTK では 1 で来ることがある。
	wxSize size = event.GetSize();
	if (size.x <= 1 || size.y <= 1) {
		return;
	}

	// Layout() を呼ぶために必要
	event.Skip();

	// XXX レイアウトが終わった後でないと DoSize() できない
//	DoSize(size);
}

// サイズ変更処理。パネルサイズと倍率が変わったら呼ぶこと。
// 倍率が変わった場合もスクロールバーの再設定とかがあるため。
void
WXPlaneVRAMPanel::DoSize(wxSize size)
{
	wxSize viewsize = viewctl->GetClientSize();
	int virtual_width = viewctl->virtual_width;
	int virtual_height = viewctl->virtual_height;
	Rect& view = viewctl->view;

	// スクロールバーが右端まで行っている場合は右を固定したように見せる。
	// ただし全体が見えている場合は除く。

	if (view.x > 0 &&
	    (view.x + view.w == virtual_width ||
	     view.x + viewsize.x >= virtual_width))
	{
		view.x = virtual_width - viewsize.x;
		if (view.x < 0) {
			view.x = 0;
		}
	}
	view.w = viewsize.x;

	// 下端も同様
	if (view.y > 0 &&
	    (view.y + view.h >= virtual_height ||
	     view.y + size.y >= virtual_height))
	{
		view.y = virtual_height - viewsize.y;
		if (view.y < 0) {
			view.y = 0;
		}
	}
	view.h = viewsize.y;

	// スクロールバー(横)を再設定
	assert(hscroll);
	int hpos = view.x;
	int hthumbsize = view.w;
	int hrange = virtual_width;
	int hpagesize = view.w;
	hscroll->SetScrollParam(hpos, hthumbsize, hrange, hpagesize);

	// スクロールバー(縦)を再設定
	assert(vscroll);
	int vpos = view.y;
	int vthumbsize = view.h;
	int vrange = virtual_height;
	int vpagesize = view.h;
	vscroll->SetScrollParam(vpos, vthumbsize, vrange, vpagesize);
}

// プレーン選択変更
void
WXPlaneVRAMPanel::EnablePlane(int plane, bool value)
{
	viewctl->EnablePlane(plane, value);
}

// パレット適用変更
void
WXPlaneVRAMPanel::EnablePalette(bool value)
{
	viewctl->EnablePalette(value);
}

// 倍率を設定
void
WXPlaneVRAMPanel::SetScale(int scale_)
{
	viewctl->SetScale(scale_);

	int vscr_width  = vscroll->GetClientSize().x;
	int hscr_height = hscroll->GetClientSize().y;
	SetMaxClientSize(wxSize(viewctl->virtual_width + vscr_width,
		viewctl->virtual_height + hscr_height));

	DoSize(GetSize());
}


//
// プレーン VRAM ウィンドウ
//

// イベントテーブル
wxBEGIN_EVENT_TABLE(WXPlaneVRAMWindow, inherited)
	EVT_COMMAND_RANGE(ID_PLANE0, ID_PLANE7, wxEVT_CHECKBOX,
		WXPlaneVRAMWindow::OnPlane)
	EVT_CHECKBOX(ID_PALETTE, WXPlaneVRAMWindow::OnApplyPalette)
	EVT_CHOICE(ID_SCALE, WXPlaneVRAMWindow::OnScale)
	EVT_TIMER(wxID_ANY, WXPlaneVRAMWindow::OnTimer)
wxEND_EVENT_TABLE()

// コンストラクタ
WXPlaneVRAMWindow::WXPlaneVRAMWindow(wxWindow *parent, const wxString& name)
	: inherited(parent, wxID_ANY, name,
		inherited::DEFAULT_STYLE | wxRESIZE_BORDER)
{
	int nplane = gPlaneVRAM->GetPlaneCount();

	auto *topsizer = new wxBoxSizer(wxVERTICAL);

	// 上にコントロールパネル(のまず下敷き)
	ctrlpanel = new WXBitmapPanel(this);
	topsizer->Add(ctrlpanel, 0, wxEXPAND);

	// コントロールパネル用の横 Sizer
	auto *ctrlbox = new wxBoxSizer(wxHORIZONTAL);

	// プレーン選択チェックボックス (#3, #2, #1, #0 の順に並べる)
	for (int i = planesw.size() - 1; i >= 0; i--) {
		planesw[i] = new wxCheckBox(ctrlpanel, ID_PLANE(i),
			string_format("#%d", i));
		ctrlbox->Add(planesw[i], 0, wxALIGN_CENTER);
	}

	// パレット合成チェックボックス
	applysw = new wxCheckBox(ctrlpanel, ID_PALETTE, _("Apply palette"));
	ctrlbox->Add(applysw, 0, wxALIGN_CENTER);

	// セパレータを出したいのだが…
	ctrlbox->Add(new wxStaticText(ctrlpanel, wxID_ANY, "|"), 0, wxALIGN_CENTER);

	// 倍率
	wxString scale_choice[] = {
		"x1",
		"x2",
		"x4",
	};
	scalesw = new wxChoice(ctrlpanel, ID_SCALE,
		wxDefaultPosition, wxDefaultSize,
		countof(scale_choice), scale_choice);
	ctrlbox->Add(scalesw, 0, wxALIGN_CENTER);

	// sizer と下敷きパネルを紐付ける
	ctrlpanel->SetSizer(ctrlbox);
	ctrlbox->SetSizeHints(ctrlpanel);

	statuspanel = new WXTextScreen(this, nnSize(50, nplane));
	topsizer->Add(statuspanel, 0, wxEXPAND);

	viewpanel = new WXPlaneVRAMPanel(this);
	topsizer->Add(viewpanel, 1, wxEXPAND);

	SetSizer(topsizer);

	// 最小サイズは上コントロール部の横幅が最小基準
	wxSize minsize = ctrlpanel->GetMinClientSize();
	minsize.y += viewpanel->GetMinClientSize().y;
	SetMinClientSize(minsize);
	// 最大サイズは DoScale() 内で設定している

	// 1bpp ならプレーン選択はすべて無効、4bpp ならすべて有効。
	if (nplane == 1) {
		for (int i = 0; i < planesw.size(); i++) {
			planesw[i]->Enable(false);
		}
	}

	// 初期値をコントロールにセットする
	if (sticky_used) {
		// 前回の値があればそれを使う。
		for (int i = 0; i < planesw.size(); i++) {
			planesw[i]->SetValue(sticky_plane[i]);
		}
		applysw->SetValue(sticky_apply);
		scalesw->SetSelection(sticky_scale);
	} else {
		if (nplane == 1) {
			// 初回 1bpp なら #0 プレーンのみオン。
			for (int i = 0; i < planesw.size(); i++) {
				planesw[i]->SetValue((i == 0));
			}
			applysw->SetValue(true);
		} else {
			// 初回 4bpp なら全プレーンオン。
			for (int i = 0; i < planesw.size(); i++) {
				planesw[i]->SetValue(true);
			}
			applysw->SetValue(true);
		}
		scalesw->SetSelection(0);
	}

	// SetValue() ではイベントが飛ばないので、
	// 直接呼び出してイベントが来たことにする。うーん…。
	for (int i = 0; i < planesw.size(); i++) {
		DoPlane(i, planesw[i]->IsChecked());
	}
	DoApplyPalette(applysw->IsChecked());
	DoScale(scalesw->GetSelection());

	// 更新タイマーはモニターパネル同様の 30Hz にしておく。
	timer.SetOwner(this);
	timer.Start(33);
}

// デストラクタ
WXPlaneVRAMWindow::~WXPlaneVRAMWindow()
{
}

void
WXPlaneVRAMWindow::FontChanged()
{
	// 親 (wxSubWindow) の FontChanged() を実行しないようにするための
	// 空のオーバーライド。親の FontChanged() が呼ぶ DoSize() が
	// ウィンドウサイズを変更しようとするが、ここはテキストフォントを
	// 持っていないのでその操作は不要。うーん。
}

// プレーン選択のチェックボックスイベント
void
WXPlaneVRAMWindow::OnPlane(wxCommandEvent& event)
{
	int plane = event.GetId() - ID_PLANE0;
	bool value = event.IsChecked();

	DoPlane(plane, value);
}

// プレーン選択 ON/OFF 処理
void
WXPlaneVRAMWindow::DoPlane(int plane, bool value)
{
	// コントロールに指示
	viewpanel->EnablePlane(plane, value);

	// 値を保存
	sticky_plane[plane] = value;
	sticky_used = true;
}

// パレット合成のチェックボックスイベント
void
WXPlaneVRAMWindow::OnApplyPalette(wxCommandEvent& event)
{
	bool value = event.IsChecked();

	DoApplyPalette(value);
}

// パレット合成 ON/OFF 処理
void
WXPlaneVRAMWindow::DoApplyPalette(bool value)
{
	// コントロールに指示
	viewpanel->EnablePalette(value);

	// 値を保存
	sticky_apply = value;
	sticky_used = true;
}

// 倍率変更イベント
void
WXPlaneVRAMWindow::OnScale(wxCommandEvent& event)
{
	int scale_index = event.GetSelection();

	DoScale(scale_index);
}

// 倍率変更処理
void
WXPlaneVRAMWindow::DoScale(int scale_index)
{
	viewpanel->SetScale(1 << scale_index);

	// 最大サイズは表示部の最大サイズから
	wxSize maxsize = viewpanel->GetMaxClientSize();
	maxsize.y += ctrlpanel->GetMinClientSize().y;
	SetMaxClientSize(maxsize);

	// 保存しておくのはインデックスのほう
	sticky_scale = scale_index;
}

// タイマーイベント
void
WXPlaneVRAMWindow::OnTimer(wxTimerEvent& event)
{
	viewpanel->Refresh();

	// この位置の情報で screen を更新
	TextScreen& screen = statuspanel->GetScreen();
	const wxPoint& cursor = viewpanel->GetCursor();

	// UpdateInfo() は cursor.x が範囲外の時の描画も行っている。
	gPlaneVRAM->UpdateInfo(screen, cursor.x, cursor.y);
	statuspanel->Refresh();
}

// static 変数
std::array<bool, WXPlaneVRAMWindow::NPLANES> WXPlaneVRAMWindow::sticky_plane;
bool WXPlaneVRAMWindow::sticky_apply;
int WXPlaneVRAMWindow::sticky_scale;
bool WXPlaneVRAMWindow::sticky_used;
